SystemOrganization addCategory: #'Files-System'!
SystemOrganization addCategory: #'Files-Kernel'!
SystemOrganization addCategory: #'Files-Directories'!

SequenceableCollection subclass: #SourceFileArray
	instanceVariableNames: 'files'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Files-System'!

!SourceFileArray commentStamp: '<historical>' prior: 0!
This is a variation on StandardSourceFileArray that provides a larger maximum changes file size.

The available address space for source pointers in a traditional CompiledMethod is 16r1000000 through 16r4FFFFFF. StandardSourceFileArray maps positions in the sources file to address range 16r1000000 through 16r1FFFFFF and 16r3000000 through 16r3FFFFFF, and positions in the changes file to address range 16r2000000 through 16r2FFFFFF and 16r4000000 through 16r4FFFFFF. This permits a maximum file size of 16r2000000 (32MB) for both the sources file and the changes file. 

This implementation extends the source pointer address space using bit 25 of the source pointer to identify the external sources and changes files, with the remaining high order bits treated as address extension. This limits the number of external file references to two (the traditional sources and changes files). If additional external file references are needed in the future, some higher order bits in the source pointer address space should be allocated for that purpose.

The use of bit 25 of the source pointer for file references permits backward compatibility with StandardSourceFileArray, with essentially unlimited address space expansion for the sources and changes files.
!

!SourceFileArray class methodsFor: 'initialize-release' stamp: 'MarcusDenker 5/3/2012 13:30'!
install
	"Replace SourceFiles by an instance of me with the standard sources and changes files.
	This only works if SourceFiles is either an Array or an instance of this class"

	"SourceFileArray install"

	^ SourceFiles := self new! !

!SourceFileArray class methodsFor: 'instance creation' stamp: 'MarcusDenker 5/3/2012 13:30'!
new: nFiles
	^self new initialize: nFiles.! !

!SourceFileArray methodsFor: 'accessing' stamp: 'MarcusDenker 5/3/2012 13:27'!
at: index
	^files at: index! !

!SourceFileArray methodsFor: 'accessing' stamp: 'MarcusDenker 5/3/2012 13:27'!
at: index put: aFile
	^ files at: index put: aFile! !

!SourceFileArray methodsFor: 'sourcepointer conversion' stamp: 'MarcusDenker 5/3/2012 13:28'!
fileIndexFromSourcePointer: anInteger
	"Return the index of the source file which contains the source chunk addressed by anInteger"

	(anInteger bitAnd: 16r1000000) ~= 0
		ifTrue: [^1 "sources file"]
		ifFalse: [anInteger >= 16r1000000
			ifTrue: [^2 "changes file"]
			ifFalse: [^0 "compatibility with StandardSourceFileArray"]]! !

!SourceFileArray methodsFor: 'sourcepointer conversion' stamp: 'MarcusDenker 5/3/2012 13:28'!
filePositionFromSourcePointer: anInteger 
	"Return the position of the source chunk addressed by anInteger"

	| hi lo |
	hi := anInteger // 33554432.
	lo := anInteger \\ 16777216.
	((anInteger bitAnd: 16777216) ~= 0
			or: [anInteger < 16777216 "compatibility with StandardSourceFileArray"])
		ifTrue: [^ hi * 16777216 + lo"sources file"]
		ifFalse: [^ hi - 1 * 16777216 + lo"changes file"]! !

!SourceFileArray methodsFor: 'initialize-release' stamp: 'MarcusDenker 5/3/2012 13:27'!
initialize
	files := Array new: 2.
	files at: 1 put: (SourceFiles at: 1).
	files at: 2 put: (SourceFiles at: 2)! !

!SourceFileArray methodsFor: 'initialize-release' stamp: 'MarcusDenker 5/3/2012 13:28'!
initialize: nFiles
	files := Array new: nFiles! !

!SourceFileArray methodsFor: 'accessing' stamp: 'MarcusDenker 5/3/2012 13:28'!
size
	^files size! !

!SourceFileArray methodsFor: 'sourcepointer conversion' stamp: 'MarcusDenker 5/3/2012 13:29'!
sourcePointerFromFileIndex: index andPosition: position
	"Return a sourcePointer encoding the given file index and position"

	| hi lo |
	(index = 1 or: [index = 2])
		ifFalse: [self error: 'invalid source file index'].
	position < 0 ifTrue: [self error: 'invalid source code pointer'].
	hi := position // 16r1000000 *2 + index.
	lo := position \\ 16r1000000.
	^ hi * 16r1000000 + lo
! !

ReadWriteStream subclass: #FileStream
	instanceVariableNames: 'rwmode'
	classVariableNames: 'Stderr Stdin StdioFiles Stdout TheStdioHandles'
	poolDictionaries: ''
	category: 'Files-Kernel'!

!FileStream commentStamp: '<historical>' prior: 0!
I represent a Stream that accesses a FilePage from a File. One use for my instance is to access larger "virtual Strings" than can be stored contiguously in main memory. I restrict the objects stored and retrieved to be Integers or Characters. An end of file pointer terminates reading; it can be extended by writing past it, or the file can be explicitly truncated.
	
To use the file system for most applications, you typically create a FileStream. This is done by sending a message to a FileDirectory (file:, oldFile:, newFile:, rename:newName:) which creates an instance of me. Accesses to the file are then done via my instance.

*** On DOS, files cannot be shortened!!  ***  To overwrite a file with a shorter one, first delete the old file (FileDirectory deleteFilePath: 'Hard Disk:aFolder:dataFolder:foo') or (aFileDirectory deleteFileNamed: 'foo').  Then write your new shorter version.!

!FileStream class methodsFor: 'concrete classes' stamp: 'yo 7/5/2004 20:18'!
concreteStream
	"Who should we really direct class queries to?  "
	^ MultiByteFileStream.
! !

!FileStream class methodsFor: 'utils' stamp: 'stephane.ducasse 7/10/2009 16:30'!
convertCRtoLF: fileName
	"Convert the given file to LF line endings. Put the result in a file with the extention '.lf'"

	| in out c justPutCR |
	in := (self readOnlyFileNamed: fileName) binary.
	out :=  (self newFileNamed: fileName, '.lf') binary.
	justPutCR := false.
	[in atEnd] whileFalse: [
		c := in next.
		c = 10
			ifTrue: [
				out nextPut: 13.
				justPutCR := true]
			ifFalse: [
				(justPutCR and: [c = 10]) ifFalse: [out nextPut: c].
				justPutCR := false]].
	in close.
	out close.
! !

!FileStream class methodsFor: 'instance creation' stamp: 'StephaneDucasse 8/3/2010 18:23'!
detectFile: aBlock do: anotherBlock

	^aBlock value
		ifNil: [nil]
      	ifNotNil: [:file| [anotherBlock value: file] ensure: [file close]]! !

!FileStream class methodsFor: 'file reader services' stamp: 'CamilloBruni 5/23/2012 16:56'!
fileIn: fullName
	"File in the entire contents of the file specified by the name provided"

	| ff fn |
	fullName ifNil: [^ Beeper beep].
	fn := fullName asFileReference.
	fn := (Smalltalk hasClassNamed: #GZipReadStream) 
		ifTrue: [(Smalltalk classNamed:  #GZipReadStream) uncompressedFileName: fn fullName]
		ifFalse: [fn fullName].
	ff := self readOnlyFileNamed: fn.
	ff fileIn.
! !

!FileStream class methodsFor: 'instance creation' stamp: ''!
fileNamed: fileName 
	^ self concreteStream fileNamed: (self fullName: fileName)! !

!FileStream class methodsFor: 'instance creation' stamp: 'VeronicaUquillas 6/11/2010 14:52'!
fileNamed: fileName do: aBlock
	"Returns the result of aBlock."
	
	^ self detectFile: [ self fileNamed: fileName ] do: aBlock! !

!FileStream class methodsFor: 'stdio' stamp: 'StephaneDucasse 5/18/2011 23:13'!
flushAndVoidStdioFiles

	StdioFiles ifNotNil: [
		StdioFiles do: [ :file |
			file ifNotNil: [ 
				file isReadOnly ifFalse: [
					[ file flush ]
						on: Error
						do: [ :ex | "care less" ] ] ] ].
		self voidStdioFiles ]! !

!FileStream class methodsFor: 'instance creation' stamp: 'tpr 10/16/2001 12:49'!
forceNewFileNamed: fileName
 	"Create a new file with the given name, and answer a stream opened for writing on that file. If the file already exists, delete it without asking before creating the new file."

	^self concreteStream forceNewFileNamed: fileName! !

!FileStream class methodsFor: 'instance creation' stamp: 'VeronicaUquillas 6/11/2010 14:52'!
forceNewFileNamed: fileName do: aBlock
	"Returns the result of aBlock."
	
	^ self detectFile: [ self forceNewFileNamed: fileName ] do: aBlock! !

!FileStream class methodsFor: 'instance creation' stamp: 'CamilloBruni 5/4/2012 21:11'!
fullName: fileName
	^ fileName asFileReference fullName ! !

!FileStream class methodsFor: 'initialize-release' stamp: 'CamilloBruni 5/9/2012 12:07'!
initialize
	TheStdioHandles := Array new: 3.
	
	"original comment was: the intent being before: AutoStart"
	Smalltalk
		addToStartUpList: self after: DiskStore; 
		addToShutDownList: self after: DiskStore! !

!FileStream class methodsFor: 'instance creation' stamp: 'TPR 8/26/1999 10:49'!
isAFileNamed: fName
	"return whether a file exists with the given name"
	^self concreteStream isAFileNamed: (self fullName: fName)! !

!FileStream class methodsFor: 'file reader services' stamp: 'tpr 9/15/2005 15:06'!
isSourceFileSuffix: suffix

	^ FileStream sourceFileSuffixes includes: suffix
! !

!FileStream class methodsFor: 'instance creation' stamp: 'di 2/15/98 14:03'!
new
	^ self basicNew! !

!FileStream class methodsFor: 'instance creation' stamp: ''!
newFileNamed: fileName 
	^ self concreteStream newFileNamed: (self fullName: fileName)! !

!FileStream class methodsFor: 'instance creation' stamp: 'VeronicaUquillas 6/11/2010 14:52'!
newFileNamed: fileName do: aBlock
	"Returns the result of aBlock."
	
	^ self detectFile: [ self newFileNamed: fileName ] do: aBlock! !

!FileStream class methodsFor: 'stdio' stamp: 'StephaneDucasse 5/18/2011 23:15'!
newForStdio
	"This is a hook for subclasses to initialize themselves properly."

	^self new! !

!FileStream class methodsFor: 'instance creation' stamp: ''!
oldFileNamed: fileName 
	^ self concreteStream oldFileNamed: (self fullName: fileName)! !

!FileStream class methodsFor: 'instance creation' stamp: 'VeronicaUquillas 6/11/2010 14:52'!
oldFileNamed: fileName do: aBlock
	"Returns the result of aBlock."
	
	^ self detectFile: [ self oldFileNamed: fileName ] do: aBlock! !

!FileStream class methodsFor: 'instance creation' stamp: 'StephaneDucasse 8/21/2010 22:50'!
oldFileOrNoneNamed: fileName
	"If the file exists, answer a read-only FileStream on it. If it doesn't, answer nil."

	| fullName |
	fullName := self fullName: fileName.
	^ (self concreteStream isAFileNamed: fullName)
		ifTrue: [ self concreteStream readOnlyFileNamed: fullName]
		ifFalse: [ nil].
! !

!FileStream class methodsFor: 'instance creation' stamp: ''!
readOnlyFileNamed: fileName 
	^ self concreteStream readOnlyFileNamed: (self fullName: fileName)! !

!FileStream class methodsFor: 'instance creation' stamp: 'StephaneDucasse 2/2/2010 21:46'!
readOnlyFileNamed: fileName do: aBlock
	"Open the existing file with the given name for read-only access and pass it as argument to aBlock.
	Returns the result of aBlock."
	
	^ self detectFile: [ self readOnlyFileNamed: fileName ] do: aBlock! !

!FileStream class methodsFor: 'file reader services' stamp: 'stephaneducasse 2/4/2006 20:32'!
removeLineFeeds: fullName
	| fileContents |
	fileContents := ((FileStream readOnlyFileNamed: fullName) wantsLineEndConversion: true) contentsOfEntireFile.
	(FileStream newFileNamed: fullName) 
		nextPutAll: fileContents;
		close.! !

!FileStream class methodsFor: 'dnd requests' stamp: 'VeronicaUquillas 6/11/2010 14:52'!
requestDropStream: dropIndex
	"Request a read-only stream for some file that was dropped onto the application"
	^self concreteStream new requestDropStream: dropIndex.! !

!FileStream class methodsFor: 'system startup' stamp: 'StephaneDucasse 5/18/2011 23:15'!
shutDown: quitting

	quitting ifTrue: [ self flushAndVoidStdioFiles ]! !

!FileStream class methodsFor: 'file reader services' stamp: 'MarcusDenker 2/14/2010 09:35'!
sourceFileSuffixes

	^#('st' 'cs')

! !

!FileStream class methodsFor: 'stdio' stamp: 'EstebanLorenzano 6/21/2012 16:25'!
standardIOStreamNamed: moniker forWrite: forWrite
	"Create if necessary and store default stdin, stdout and other files based on the their names"
	
	| index |
	
	self flag: #todo. "This is an ugly hack, while waiting for a real fix for windows. There  several problems with this approach, but it allow us to run tests, etc."
	OSPlatform isWin32 
		ifTrue: [ ^MultiByteFileStream forceNewFileNamed: moniker asString ]. 	
	
	index := #(stdin stdout stderr) identityIndexOf: moniker.
	^((StdioFiles ifNil: [ StdioFiles := Array new: 3 ]) at: index)
		ifNil: [
			StdioFiles
				at: index 
				put: (
					(TheStdioHandles at: index)
						ifNil: [ ^self error: moniker, ' is unavailable' ]
						ifNotNil: [ :handle |
							MultiByteFileStream newForStdio
								openOnHandle: handle
								name: moniker
								forWrite: forWrite ]) ]
! !

!FileStream class methodsFor: 'system startup' stamp: 'StephaneDucasse 5/18/2011 23:19'!
startUp: resuming
	
	resuming ifTrue: [
		self voidStdioFiles.
		[ TheStdioHandles := self stdioHandles ]
			on: Error
			do: [:ex|
				TheStdioHandles isArray ifFalse: [
					TheStdioHandles := Array new: 3 ] ] ]! !

!FileStream class methodsFor: 'stdio' stamp: 'StephaneDucasse 5/18/2011 23:19'!
stderr

	^Stderr ifNil: [ Stderr := self standardIOStreamNamed: #stderr forWrite: true ]! !

!FileStream class methodsFor: 'stdio' stamp: 'SeanDeNigris 5/3/2012 22:37'!
stdin

	^Stdin ifNil: [
		Stdin := self standardIOStreamNamed: #stdin forWrite: false.
		Stdin
			disableReadBuffering;
			yourself ].! !

!FileStream class methodsFor: 'stdio' stamp: 'StephaneDucasse 5/18/2011 23:19'!
stdioHandles
	<primitive: 'primitiveFileStdioHandles' module: 'FilePlugin' error: ec>
	self primitiveFailed! !

!FileStream class methodsFor: 'stdio' stamp: 'StephaneDucasse 5/18/2011 23:20'!
stdout

	^Stdout ifNil: [ Stdout := self standardIOStreamNamed: #stdout forWrite: true ]! !

!FileStream class methodsFor: 'initialize-release' stamp: 'GabrielOmarCotelli 6/4/2009 20:36'!
unload

	FileServices unregisterFileReader: self ! !

!FileStream class methodsFor: 'stdio' stamp: 'StephaneDucasse 5/18/2011 23:20'!
voidStdioFiles

	 Stdin := Stdout := Stderr := StdioFiles := nil! !

!FileStream class methodsFor: 'file reader services' stamp: 'CamilloBruni 5/10/2012 16:09'!
writeSourceCodeFrom: aStream baseName: baseName isSt: stOrCsFlag

	| extension converter fileName |
	extension :=  stOrCsFlag ifTrue: ['.st']  ifFalse: ['.cs'].
	converter := aStream contents isAsciiString 
		ifTrue: [MacRomanTextConverter new] 
		ifFalse: [UTF8TextConverter new].
	fileName := baseName, extension.
	fileName := FileSystem disk checkName: fileName fixErrors: true.
	[FileStream newFileNamed: fileName do: [:fileStream |
		(converter isMemberOf: UTF8TextConverter)
			ifTrue: [fileStream binary.
					UTF8TextConverter writeBOMOn: fileStream].
		fileStream
			text;
			converter: converter;
			nextPutAll: aStream contents;
			close]] on: Abort do: [:e | ]! !

!FileStream methodsFor: 'converting' stamp: 'tk 2/4/2000 09:16'!
asBinaryOrTextStream
	"I can switch between binary and text data"

	^ self! !

!FileStream methodsFor: 'file modes' stamp: 'jm 9/21/1998 13:01'!
ascii
	"Set this file to ascii (text) mode."

	self subclassResponsibility
! !

!FileStream methodsFor: 'testing' stamp: 'CamilloBruni 8/1/2012 16:00'!
atEnd
	"Answer true if the current position is >= the end of file position."

	self subclassResponsibility! !

!FileStream methodsFor: 'file modes' stamp: 'jm 9/21/1998 12:59'!
binary
	"Set this file to binary mode."

	self subclassResponsibility
! !

!FileStream methodsFor: 'file open/close' stamp: 'jm 9/21/1998 13:02'!
close
	"Close this file."

	self subclassResponsibility
! !

!FileStream methodsFor: 'file open/close' stamp: 'jm 9/21/1998 13:02'!
closed
	"Answer true if this file is closed."

	self subclassResponsibility
! !

!FileStream methodsFor: 'accessing' stamp: 'stephaneducasse 2/4/2006 20:31'!
contents
	"Return the contents of the receiver. Do not close or otherwise touch the receiver. Return data in whatever mode the receiver is in (e.g., binary or text)."
	| s savePos |
	savePos := self position.
	self position: 0.
	s := self next: self size.
	self position: savePos.
	^s! !

!FileStream methodsFor: 'accessing' stamp: 'stephaneducasse 2/4/2006 20:31'!
contentsOfEntireFile
	"Read all of the contents of the receiver."

	| s binary |
	self readOnly.
	binary := self isBinary.
	self reset.	"erases knowledge of whether it is binary"
	binary ifTrue: [self binary].
	s := self next: self size.
	self close.
	^s! !

!FileStream methodsFor: 'remote file compatibility' stamp: 'RAA 9/24/2000 18:00'!
dataIsValid

	self flag: #bob.		"we needed this if a remote stream, but could be local as well"! !

!FileStream methodsFor: 'accessing' stamp: 'abc 5/11/2012 23:29'!
directoryEntry
	^(self directory / self localName) entry! !

!FileStream methodsFor: 'file accessing' stamp: 'CamilloBruni 8/1/2012 16:05'!
file
	"Answer the file for the page the receiver is streaming over."

	self subclassResponsibility! !

!FileStream methodsFor: 'file open/close' stamp: 'jm 9/21/1998 13:03'!
flush
	"When writing, flush the current buffer out to disk."

	self subclassResponsibility
! !

!FileStream methodsFor: 'file accessing' stamp: 'CamilloBruni 5/7/2012 02:09'!
localName

	^ self name asFileReference basename
! !

!FileStream methodsFor: 'printing' stamp: 'tk 12/5/2001 09:12'!
longPrintOn: aStream
	"Do nothing, so it will print short.  Called to print the error file.  If the error was in a file operation, we can't read the contents of that file.  Just print its name instead."
! !

!FileStream methodsFor: 'printing' stamp: 'tk 12/5/2001 09:32'!
longPrintOn: aStream limitedTo: sizeLimit indent: indent

	"Do nothing, so it will print short.  Called to print the error file.  If the error was in a file operation, we can't read the contents of that file.  Just print its name instead."

	aStream cr! !

!FileStream methodsFor: 'accessing' stamp: 'CamilloBruni 5/23/2012 16:55'!
mimeTypes
	^ self name asFileReference mimeTypes.! !

!FileStream methodsFor: 'file accessing' stamp: 'CamilloBruni 8/1/2012 16:11'!
name
	"Answer the name of the file for the page the receiver is streaming over."

	self subclassResponsibility! !

!FileStream methodsFor: 'accessing' stamp: 'stephaneducasse 2/4/2006 20:31'!
next

	(position >= readLimit and: [self atEnd])
		ifTrue: [^nil]
		ifFalse: [^collection at: (position := position + 1)]! !

!FileStream methodsFor: 'accessing' stamp: 'nice 11/22/2009 18:11'!
next: anInteger

	| newCollection howManyRead increment |
	newCollection := self collectionSpecies new: anInteger.
	howManyRead := 0.
	[howManyRead < anInteger] whileTrue:
		[self atEnd ifTrue:
			[(howManyRead + 1) to: anInteger do: [:i | newCollection at: i put: (self next)].
			^newCollection].
		increment := (readLimit - position) min: (anInteger - howManyRead).
		newCollection replaceFrom: (howManyRead + 1)
			to: (howManyRead := howManyRead + increment)
			with: collection
			startingAt: (position + 1).
		position := position + increment].
	^newCollection! !

!FileStream methodsFor: 'accessing' stamp: ''!
nextPut: aByte
	"1/31/96 sw: subclassResponsibility"

	self subclassResponsibility! !

!FileStream methodsFor: 'accessing' stamp: ''!
nextPutAll: aCollection
	"1/31/96 sw: made subclass responsibility"

	self subclassResponsibility! !

!FileStream methodsFor: 'positioning' stamp: 'CamilloBruni 8/1/2012 16:12'!
position
	"Answer the current character position in the file."

	self subclassResponsibility! !

!FileStream methodsFor: 'positioning' stamp: 'CamilloBruni 8/1/2012 16:13'!
position: pos
	"Set the current character position in the file to pos."

	self subclassResponsibility! !

!FileStream methodsFor: 'printing' stamp: ''!
printOn: aStream

	super printOn: aStream.
	aStream nextPutAll: ' on '.
	self file printOn: aStream! !

!FileStream methodsFor: 'file modes' stamp: 'jm 9/21/1998 12:59'!
readOnly
	"Set this file's mode to read-only."

	self subclassResponsibility
! !

!FileStream methodsFor: 'file modes' stamp: 'mir 8/24/2004 17:58'!
readOnlyStream
	^self readOnly! !

!FileStream methodsFor: 'file modes' stamp: 'jm 9/21/1998 13:00'!
readWrite
	"Set this file's mode to read-write."

	self subclassResponsibility
! !

!FileStream methodsFor: 'file open/close' stamp: 'jm 9/21/1998 13:04'!
reopen
	"Ensure that the receiver is open, re-open it if necessary."
	"Details: Files that were open when a snapshot occurs are no longer valid when the snapshot is resumed. This operation re-opens the file if that has happened."

	self subclassResponsibility
! !

!FileStream methodsFor: 'positioning' stamp: 'CamilloBruni 8/1/2012 16:15'!
reset
	"Set the current character position to the beginning of the file."

	self subclassResponsibility! !

!FileStream methodsFor: 'positioning' stamp: 'CamilloBruni 8/1/2012 16:16'!
setToEnd
	"Set the current character position to the end of the File. The same as
	self position: self size."

	self subclassResponsibility! !

!FileStream methodsFor: 'accessing' stamp: 'CamilloBruni 8/1/2012 16:16'!
size
	"Answer the size of the file in characters."

	self subclassResponsibility! !

!FileStream methodsFor: 'positioning' stamp: 'CamilloBruni 8/1/2012 16:16'!
skip: n
	"Set the character position to n characters from the current position.
	Error if not enough characters left in the file."
 
	self subclassResponsibility! !

!FileStream methodsFor: 'file modes' stamp: 'jm 9/21/1998 13:01'!
text
	"Set this file to text (ascii) mode."

	self ascii.
! !

!FileStream methodsFor: 'positioning' stamp: 'JMM 5/24/2001 22:58'!
truncate: pos
	"Truncate file to pos"

	self subclassResponsibility! !

FileStream subclass: #StandardFileStream
	instanceVariableNames: 'name fileID buffer1'
	classVariableNames: 'Registry'
	poolDictionaries: ''
	category: 'Files-Kernel'!

!StandardFileStream commentStamp: 'ul 12/6/2009 05:13' prior: 0!
Provides a simple, platform-independent, interface to a file system. The instance variable rwmode, inherited from class PositionableStream, here is used to hold a Boolean -- true means opened for read-write, false means opened for read-only.  2/12/96 sw

I implement a simple read buffering scheme with the variables defined in PositionableStream (which are unused in me otherwise) in the following way:
	collection	<ByteString> or <ByteArray>	This is the buffer.
	position	<Integer>	The relative position in the buffer. Greater or equal to zero.
	readLimit	<Integer>	The number of bytes buffered. Greater or equal to zero.
Read buffering is enabled with #enableReadBuffering, disabled with #disableReadBuffering and it is enabled by default. The buffer is filled when a read attempt of an unbuffered absolute position is requested, or when a negative repositioning is made (with #position: with an argument < than the current absolute position) to an absolute position which is not buffered. In the first case, the buffer is positioned to the given absolute position. In the latter case the repositioning is made to the requested absolute position minus fourth of the buffer size. This means that further small negative repositionings won't result in buffer flushing. This is really useful when filing in code.
The read buffer is flushed (#flushReadBuffer) whenever a write attempt is made.
The buffer state is valid if and only if collection is not nil and position < readLimit.!

StandardFileStream subclass: #MultiByteFileStream
	instanceVariableNames: 'converter lineEndConvention wantsLineEndConversion'
	classVariableNames: 'Cr CrLf Lf LineEndDefault LineEndStrings LookAheadCount'
	poolDictionaries: ''
	category: 'Files-Kernel'!

!MultiByteFileStream commentStamp: '<historical>' prior: 0!
The central class to access the external file.  The interface of this object is similar to good old StandardFileStream, but internally it asks the converter, which is a sub-instance of TextConverter, and do the text conversion.

  It also combined the good old CrLfFileStream.  CrLfFileStream class>>new now returns an instance of MultiByteFileStream.

  There are several pitfalls:

  * You always have to be careful about the binary/text distinction.  In #text mode, it usually interpret the bytes.
  * A few file pointer operations treat the file as uninterpreted byte no matter what.  This means that if you use 'fileStream skip: -1', 'fileStream position: x', etc. in #text mode, the file position can be in the middle of multi byte character.  If you want to implement some function similar to #peek for example, call the saveStateOf: and restoreStateOf: methods to be able to get back to the original state.
  * #lineEndConvention: and #wantsLineEndConversion: (and #binary) can cause some puzzling situation because the inst var lineEndConvention and wantsLineEndConversion are mutated.  If you have any suggestions to clean up the protocol, please let me know.!

!MultiByteFileStream class methodsFor: 'class initialization' stamp: 'yo 2/21/2004 02:45'!
defaultToCR

	"MultiByteFileStream defaultToCR"
	LineEndDefault := #cr.
! !

!MultiByteFileStream class methodsFor: 'class initialization' stamp: 'yo 2/21/2004 02:45'!
defaultToCRLF

	"MultiByteFileStream defaultToCRLF"
	LineEndDefault := #crlf.! !

!MultiByteFileStream class methodsFor: 'class initialization' stamp: 'yo 2/21/2004 02:46'!
defaultToLF

	"MultiByteFileStream defaultToLF"
	LineEndDefault := #lf.
! !

!MultiByteFileStream class methodsFor: 'class initialization' stamp: 'CamilloBruni 5/7/2012 11:30'!
guessDefaultLineEndConvention
	"Lets try to guess the line end convention from what we know about the
	path name delimiter from FileDirectory."
	
	FileSystem disk delimiter = $:
		ifTrue: [^ self defaultToCR].
		
	FileSystem disk delimiter = $/
		ifTrue: [^ (OSPlatform isMacOSX or: [OSPlatform isUnix])
				ifTrue: [ self defaultToLF]
				ifFalse: [ self defaultToCR]].
			
	FileSystem disk delimiter = $\
		ifTrue: [^ self defaultToCRLF].
		
	"in case we don't know"
	^ self defaultToCR! !

!MultiByteFileStream class methodsFor: 'class initialization' stamp: 'yo 2/21/2004 02:44'!
initialize

	"MultiByteFileStream initialize"
	Cr := Character cr.
	Lf := Character lf.
	CrLf := String with: Cr with: Lf.
	LineEndStrings := Dictionary new.
	LineEndStrings at: #cr put: (String with: Character cr).
	LineEndStrings at: #lf put: (String with: Character lf).
	LineEndStrings at: #crlf put: (String with: Character cr with: Character lf).
	LookAheadCount := 2048.
	Smalltalk addToStartUpList: self.
	self startUp.
! !

!MultiByteFileStream class methodsFor: 'accessing' stamp: 'StephaneDucasse 10/18/2010 14:05'!
lineEndDefault
	"Answer the default line-ending convention that will be used by default, which was determined at start-up by looking at platform attributes."
	^ LineEndDefault
! !

!MultiByteFileStream class methodsFor: 'instance creation' stamp: 'yo 8/28/2002 11:43'!
newFrom: aFileStream

	| rw n |
	n := aFileStream name.
	rw := aFileStream isReadOnly not.
	aFileStream close.
	^self new open: n forWrite: rw.
! !

!MultiByteFileStream class methodsFor: 'class initialization' stamp: 'yo 2/21/2004 02:44'!
startUp

	self guessDefaultLineEndConvention.
! !

!MultiByteFileStream methodsFor: 'remnant' stamp: 'yo 8/28/2002 11:06'!
accepts: aSymbol

 	^ converter accepts: aSymbol.
! !

!MultiByteFileStream methodsFor: 'accessing' stamp: 'yo 2/21/2004 02:57'!
ascii

	super ascii.
	self detectLineEndConvention.
! !

!MultiByteFileStream methodsFor: 'public' stamp: 'ClementBera 11/2/2012 12:01'!
back

       ^converter backFromStream: self! !

!MultiByteFileStream methodsFor: 'crlf private' stamp: 'yo 2/24/2004 13:38'!
bareNext

	 ^ self converter nextFromStream: self.
! !

!MultiByteFileStream methodsFor: 'private basic' stamp: 'ClementBera 11/2/2012 11:59'!
basicBack

       self position = 0 ifTrue: [ self error: 'CantGoBack' ].
       self skip: -1.
       ^self basicPeek
! !

!MultiByteFileStream methodsFor: 'fileIn/Out' stamp: 'HenrikSperreJohansen 2/9/2010 11:23'!
basicChunk
	"If our buffer in collection contains an chunk with no embedded !!'s, nor
	any non-ascii characters, return that.
	This presumes the source code encoding is unambiguously ascii-compatible"
	| bufferIX goodString |
	"Not possible if read buffering disabled"
	collection ifNil: [^nil].
	^ ((bufferIX := (collection indexOf: $!! startingAt: position + 1) min: readLimit +1) > 0
			and: [bufferIX < collection size
					and: [(collection at: bufferIX + 1)
								~= $!!
							and: [goodString := collection copyFrom: position + 1 to: bufferIX - 1.
								goodString isAsciiString]]])
		ifTrue: [ position := bufferIX.
				goodString]! !

!MultiByteFileStream methodsFor: 'private basic' stamp: 'md 10/17/2004 16:09'!
basicNext: anInteger

	^ super next: anInteger.
! !

!MultiByteFileStream methodsFor: 'private basic' stamp: 'yo 8/28/2002 11:07'!
basicNext: n into: aString

	^ super next: n into: aString.
! !

!MultiByteFileStream methodsFor: 'private basic' stamp: 'HenrikSperreJohansen 8/17/2012 11:30'!
basicNext: anInteger putAll: aCollection startingAt: startIndex


        ^super next: anInteger putAll: aCollection startingAt: startIndex! !

!MultiByteFileStream methodsFor: 'private basic' stamp: 'yo 8/28/2002 11:07'!
basicNextInto: aString

	^ super nextInto: aString.
! !

!MultiByteFileStream methodsFor: 'private basic' stamp: 'yo 8/28/2002 11:07'!
basicNextPut: char

	^ super nextPut: char.
! !

!MultiByteFileStream methodsFor: 'private basic' stamp: 'yo 8/28/2002 11:07'!
basicNextPutAll: aString

	^ super nextPutAll: aString.
! !

!MultiByteFileStream methodsFor: 'private basic' stamp: 'yo 8/28/2002 11:07'!
basicPeek

	^ super peek
! !

!MultiByteFileStream methodsFor: 'private basic' stamp: 'yo 8/28/2002 11:08'!
basicPosition

	^ super position.
! !

!MultiByteFileStream methodsFor: 'private basic' stamp: 'yo 8/28/2002 11:08'!
basicPosition: pos

	^ super position: pos.
! !

!MultiByteFileStream methodsFor: 'private basic' stamp: 'yo 8/28/2002 11:08'!
basicReadInto: byteArray startingAt: startIndex count: count

	^ super readInto: byteArray startingAt: startIndex count: count.
! !

!MultiByteFileStream methodsFor: 'private basic' stamp: 'yo 8/28/2002 11:08'!
basicSetToEnd

	^ super setToEnd.
! !

!MultiByteFileStream methodsFor: 'private basic' stamp: 'yo 8/28/2002 11:08'!
basicSkip: n

	^ super skip: n.
! !

!MultiByteFileStream methodsFor: 'private basic' stamp: 'yo 8/28/2002 11:08'!
basicUpTo: delim

	^ super upTo: delim.
! !

!MultiByteFileStream methodsFor: 'private basic' stamp: 'yo 8/28/2002 11:09'!
basicVerbatim: aString

	^ super verbatim: aString.
! !

!MultiByteFileStream methodsFor: 'accessing' stamp: 'nice 5/10/2009 00:14'!
binary

	super binary.
	self lineEndConvention: nil! !

!MultiByteFileStream methodsFor: 'crlf private' stamp: 'yo 2/21/2004 02:56'!
convertStringFromCr: aString 
	| inStream outStream |
	lineEndConvention ifNil: [^ aString].
	lineEndConvention == #cr ifTrue: [^ aString].
	lineEndConvention == #lf ifTrue: [^ aString copy replaceAll: Cr with: Lf].
	"lineEndConvention == #crlf"
	inStream := aString readStream.
	outStream :=  (String new: aString size) writeStream.
	[inStream atEnd]
		whileFalse: 
			[outStream nextPutAll: (inStream upTo: Cr).
			(inStream atEnd not or: [aString last = Cr])
				ifTrue: [outStream nextPutAll: CrLf]].
	^ outStream contents! !

!MultiByteFileStream methodsFor: 'crlf private' stamp: 'yo 2/21/2004 02:56'!
convertStringToCr: aString 
	| inStream outStream |
	lineEndConvention ifNil: [^ aString].
	lineEndConvention == #cr ifTrue: [^ aString].
	lineEndConvention == #lf ifTrue: [^ aString copy replaceAll: Lf with: Cr].
	"lineEndConvention == #crlf"
	inStream := aString readStream.
	outStream := (String new: aString size) writeStream.
	[inStream atEnd]
		whileFalse: 
			[outStream nextPutAll: (inStream upTo: Cr).
			(inStream atEnd not or: [aString last = Cr])
				ifTrue: 
					[outStream nextPut: Cr.
					inStream peek = Lf ifTrue: [inStream next]]].
	^ outStream contents! !

!MultiByteFileStream methodsFor: 'accessing' stamp: 'nice 5/10/2009 00:17'!
converter

	converter ifNil: [self converter: TextConverter defaultSystemConverter].
	^ converter
! !

!MultiByteFileStream methodsFor: 'accessing' stamp: 'nice 5/10/2009 00:18'!
converter: aConverter

	converter := aConverter.
	self installLineEndConventionInConverter
! !

!MultiByteFileStream methodsFor: 'crlf private' stamp: 'tbn 5/10/2012 15:29'!
detectLineEndConvention
	"Detect the line end convention used in this stream. The result may be either #cr, #lf or #crlf."
	| char numRead state |
	self isBinary ifTrue: [^ self error: 'Line end conventions are not used on binary streams'].
	wantsLineEndConversion == true ifFalse: [self lineEndConvention: nil.
					^lineEndConvention].
	self closed ifTrue: [self lineEndConvention: LineEndDefault.
					^lineEndConvention].

	"Default if nothing else found"
	numRead := 0.
	state := self converter saveStateOf: self.
	lineEndConvention := nil.
	[super atEnd not and: [numRead < LookAheadCount]]
		whileTrue: 
			[char := self next.
			char = Lf
				ifTrue: 
					[converter restoreStateOf: self with: state.
					self lineEndConvention: #lf.
					^lineEndConvention].
			char = Cr
				ifTrue: 
					[self peek = Lf
						ifTrue: [self lineEndConvention: #crlf]
						ifFalse: [self lineEndConvention: #cr].
					converter restoreStateOf: self with: state.
					^ lineEndConvention].
			numRead := numRead + 1].
	converter restoreStateOf: self with: state.
	self lineEndConvention: LineEndDefault.
	^ lineEndConvention! !

!MultiByteFileStream methodsFor: 'crlf private' stamp: 'ul 11/25/2009 01:22'!
doConversion

	^wantsLineEndConversion == true and: [ lineEndConvention notNil ]! !

!MultiByteFileStream methodsFor: 'accessing' stamp: 'yo 8/6/2003 11:56'!
fileInEncodingName: aString

	self converter: (TextConverter newForEncoding: aString).
	super fileIn.
! !

!MultiByteFileStream methodsFor: 'private' stamp: 'ul 11/25/2009 01:28'!
installLineEndConventionInConverter

	converter ifNotNil: [
		converter installLineEndConvention: (
			(wantsLineEndConversion == true and: [ lineEndConvention notNil ]) "#doConversion is inlined here"
				ifTrue: [ LineEndStrings at: lineEndConvention ]
				ifFalse: [ nil ]) ]! !

!MultiByteFileStream methodsFor: 'accessing' stamp: 'nk 9/5/2004 12:57'!
lineEndConvention

	^lineEndConvention! !

!MultiByteFileStream methodsFor: 'accessing' stamp: 'nice 11/28/2009 13:06'!
lineEndConvention: aSymbol

	(lineEndConvention := aSymbol) ifNotNil: [wantsLineEndConversion := true].
	self installLineEndConventionInConverter! !

!MultiByteFileStream methodsFor: 'public' stamp: 'ul 12/3/2009 01:33'!
next

	| char secondChar state |
	char := (converter ifNil: [ self converter ]) nextFromStream: self.
	(wantsLineEndConversion == true and: [ lineEndConvention notNil ]) "#doConversion is inlined here"
		 ifTrue: [
			char == Cr ifTrue: [
				state := converter saveStateOf: self.
				secondChar := self bareNext.
				secondChar ifNotNil: [
					secondChar == Lf ifFalse: [ converter restoreStateOf: self with: state ] ].
				^Cr ].
			char == Lf ifTrue: [
				^Cr ] ].
	^char.

! !

!MultiByteFileStream methodsFor: 'public' stamp: 'ul 11/25/2009 01:28'!
next: anInteger 

	| multiString |
	self isBinary ifTrue: [^ super next: anInteger].
	multiString := String new: anInteger.
	1 to: anInteger do: [:index |
		| character |
		(character := self next)
			ifNotNil: [ multiString at: index put: character ]
			ifNil: [
				multiString := multiString copyFrom: 1 to: index - 1.
				(wantsLineEndConversion == true and: [ lineEndConvention notNil ]) "#doConversion is inlined here"
					ifFalse: [ ^multiString ].
				^self next: anInteger innerFor: multiString ] ].
	(wantsLineEndConversion == true and: [ lineEndConvention notNil ]) "#doConversion is inlined here"
		 ifFalse: [ ^multiString ].

	multiString := self next: anInteger innerFor: multiString.
	(multiString size = anInteger or: [self atEnd]) ifTrue: [ ^ multiString].
	^ multiString, (self next: anInteger - multiString size).
! !

!MultiByteFileStream methodsFor: 'crlf private' stamp: 'yo 2/24/2004 13:44'!
next: n innerFor: aString

	| peekChar state |
	"if we just read a CR, and the next character is an LF, then skip the LF"
	aString size = 0 ifTrue: [^ aString].
	(aString last = Character cr) ifTrue: [
		state := converter saveStateOf: self.
		peekChar := self bareNext.		"super peek doesn't work because it relies on #next"
		(peekChar notNil and: [peekChar ~= Character lf]) ifTrue: [
			converter restoreStateOf: self with: state.
		].
	].
 
	^ aString withSqueakLineEndings.
! !

!MultiByteFileStream methodsFor: 'public' stamp: 'HenrikSperreJohansen 8/17/2012 11:36'!
next: anInteger putAll: aCollection startingAt: startIndex


        (self isBinary or: [ aCollection class == ByteArray ]) ifTrue: [
                ^super next: anInteger putAll: aCollection startingAt: startIndex ].
        ^self converter next: anInteger putAll: aCollection startingAt: startIndex toStream: self! !

!MultiByteFileStream methodsFor: 'fileIn/Out' stamp: 'JohanBrichau 6/11/2010 15:34'!
nextChunk
	"Answer the contents of the receiver, up to the next terminator
	character. Doubled terminators indicate an embedded terminator
	character."
	| bufferIX goodString |
	self skipSeparators.
	^ self
		parseLangTagFor: (self basicChunk
				ifNil: [String
						new: 1000
						streamContents: [:stream | 
							| character state |
							[(character := self next) == nil
								or: [character == $!!
										and: [state := converter saveStateOf: self..
											self next ~~ $!!]]]
								whileFalse: [stream nextPut: character].
							character
								ifNotNil: [converter restoreStateOf: self with: state.]]])! !

!MultiByteFileStream methodsFor: 'public' stamp: 'yo 2/21/2004 03:26'!
nextDelimited: terminator

	| out ch save |
	out := (String new: 1000) writeStream.
	self atEnd ifTrue: [^ ''].
	save := converter saveStateOf: self.

	self next = terminator ifFalse: [
		"absorb initial terminator"
		converter restoreStateOf: self with: save.
	].
	[(ch := self next) == nil] whileFalse: [
		(ch = terminator) ifTrue: [
			self peek = terminator ifTrue: [
				self next.  "skip doubled terminator"
			] ifFalse: [
				^ out contents  "terminator is not doubled; we're done!!"
			].
		].
		out nextPut: ch.
	].
	^ out contents.
! !

!MultiByteFileStream methodsFor: 'public' stamp: 'yo 8/28/2002 11:13'!
nextMatchAll: aColl

    | save |
    save := converter saveStateOf: self.
    aColl do: [:each |
       (self next) = each ifFalse: [
            converter restoreStateOf: self with: save.
            ^ false.
		].
	].
    ^ true.
! !

!MultiByteFileStream methodsFor: 'public' stamp: 'LukasRenggli 5/8/2010 18:41'!
nextPut: aCharacter
	aCharacter isInteger 
		ifTrue: [ ^ super nextPut: aCharacter ].
	(wantsLineEndConversion == true and: [ lineEndConvention notNil ]) "#doConversion is inlined here"
		 ifTrue: [
			aCharacter = Cr
				ifTrue: [ converter nextPutAll: (LineEndStrings at: lineEndConvention) toStream: self ]
				ifFalse: [ converter nextPut: aCharacter toStream: self ].
			^aCharacter ].
	^ self converter nextPut: aCharacter toStream: self
! !

!MultiByteFileStream methodsFor: 'public' stamp: 'LukasRenggli 5/8/2010 18:41'!
nextPutAll: aCollection
	^ (self isBinary or: [ aCollection class == ByteArray ])
		ifTrue: [ super nextPutAll: aCollection ]
		ifFalse: [ self converter nextPutAll: aCollection toStream: self ]! !

!MultiByteFileStream methodsFor: 'open/close' stamp: 'HenrikSperreJohansen 11/20/2009 15:13'!
open: fileName forWrite: writeMode 
	| result |
	result := super open: fileName forWrite: writeMode.
	result ifNotNil: [
			converter ifNil: [self converter: UTF8TextConverter new].
			lineEndConvention ifNil: [ self detectLineEndConvention ]
	].
	^result! !

!MultiByteFileStream methodsFor: 'public' stamp: 'yo 2/21/2004 04:00'!
peek
	"Answer what would be returned if the message next were sent to the receiver. If the receiver is at the end, answer nil.  "

	| next save |
	self atEnd ifTrue: [^ nil].
	save := converter saveStateOf: self.
	next := self next.
	converter restoreStateOf: self with: save.
	^ next.

! !

!MultiByteFileStream methodsFor: 'public' stamp: 'CamilloBruni 9/5/2012 11:26'!
peekFor: item 

	| next state |
	state := converter saveStateOf: self.
	(next := self next) == nil ifTrue: [^ false].
	item = next ifTrue: [^ true].
	converter restoreStateOf: self with: state.
	^ false.
! !

!MultiByteFileStream methodsFor: 'private' stamp: 'StephaneDucasse 5/3/2010 22:59'!
requestDropStream: dropIndex
	"Needs to install proper converter"
	
	| result |
	result := super requestDropStream: dropIndex.
	result ifNotNil: [
		converter ifNil: [self converter: UTF8TextConverter new].
		lineEndConvention ifNil: [ self detectLineEndConvention] ].
	^result! !

!MultiByteFileStream methodsFor: 'open/close' stamp: 'nice 5/10/2009 00:18'!
reset

	super reset.
	converter ifNil: [
		self converter: UTF8TextConverter new.
	].
! !

!MultiByteFileStream methodsFor: 'private' stamp: 'nice 1/18/2010 13:42'!
setConverterForCode

	| current |
	(SourceFiles at: 2)
		ifNotNil: [self fullName = (SourceFiles at: 2) fullName ifTrue: [^ self]].
	current := self converter saveStateOf: self.
	self position: 0.
	self binary.
	((self next: 3) = #[ 16rEF 16rBB 16rBF ]) ifTrue: [
		self converter: UTF8TextConverter new
	] ifFalse: [
		self converter: MacRomanTextConverter new.
	].
	converter restoreStateOf: self with: current.
	self text.
! !

!MultiByteFileStream methodsFor: 'public' stamp: 'ul 12/7/2009 23:35'!
skipSeparators

	| state character |
	[ 
		state := converter saveStateOf: self.
		(character := self next) 
			ifNil: [ false ]
			ifNotNil: [ character isSeparator ] ] whileTrue.
	character ifNotNil: [ 
		converter restoreStateOf: self with: state ]! !

!MultiByteFileStream methodsFor: 'public' stamp: 'ul 12/8/2009 00:05'!
skipSeparatorsAndPeekNext
	"Same as #skipSeparators, but returns the next character after the separators if such exists."

	| state character |
	[ 
		state := converter saveStateOf: self.
		(character := self next) 
			ifNil: [ false ]
			ifNotNil: [ character isSeparator ] ] whileTrue.
	character ifNotNil: [ 
		converter restoreStateOf: self with: state.
		^character ].
	! !

!MultiByteFileStream methodsFor: 'public' stamp: 'ul 12/4/2009 19:08'!
upTo: delimiter

	^self collectionSpecies new: 1000 streamContents: [ :stream |
		| ch |
		[ (ch := self next) == nil or: [ ch = delimiter ] ] 
			whileFalse: [ stream nextPut: ch ] ]
! !

!MultiByteFileStream methodsFor: 'public' stamp: 'nice 12/7/2009 08:26'!
upToAnyOf: delimiters do: aBlock

	^self collectionSpecies new: 1000 streamContents: [ :stream |
		| ch |
		[ (ch := self next) == nil or: [ (delimiters includes: ch) and: [aBlock value: ch. true] ] ] 
			whileFalse: [ stream nextPut: ch ] ]! !

!MultiByteFileStream methodsFor: 'public' stamp: 'ul 12/6/2009 04:18'!
upToEnd

	^self collectionSpecies
		new: self size - self position
		streamContents: [ :stream |
			| element |
			[ (element := self next) == nil ] whileFalse: [
				stream nextPut: element ] ]
! !

!MultiByteFileStream methodsFor: 'crlf private' stamp: 'ul 11/25/2009 01:22'!
wantsLineEndConversion

	^wantsLineEndConversion == true
! !

!MultiByteFileStream methodsFor: 'remnant' stamp: 'kph 3/1/2009 15:50'!
wantsLineEndConversion: aBoolean
	
	wantsLineEndConversion :=  aBoolean.

	lineEndConvention ifNil: [ self detectLineEndConvention ]. ! !

!StandardFileStream class methodsFor: 'error handling' stamp: 'CamilloBruni 5/7/2012 02:10'!
fileDoesNotExistUserHandling: fullFileName

	| selection newName |
	selection := UIManager default chooseFrom: {
		'create a new file' translated.
		'choose another name' translated.
		'cancel' translated
	} title: (fullFileName asFileReference basename) , ' does not exist.'.
	selection = 1 ifTrue:
		[^ self new open: fullFileName forWrite: true].
	selection = 2 ifTrue:
		[ newName := UIManager default request: 'Enter a new file name'
						initialAnswer:  fullFileName.
		^ self oldFileNamed:
			(self fullName: newName)].! !

!StandardFileStream class methodsFor: 'error handling' stamp: 'SeanDeNigris 7/12/2012 08:44'!
fileExistsUserHandling: fullFileName
	| file localName choice newName newFullFileName |
	
	file := fullFileName asFileReference.
	localName := file basename.
	choice := UIManager default chooseFrom:{
		'overwrite that file'.
		'choose another name'.
		'cancel'
	} title: localName, ' already exists.'.
 
	choice = 1 ifTrue: [
		file delete.
		file exists ifTrue:  [self error: 'Could not delete the old version of that file'].
		^ self new open: fullFileName forWrite: true].

	choice = 2 ifTrue: [
		newName := UIManager default request: 'Enter a new file name' initialAnswer: fullFileName.
		newName isNil ifTrue: [Abort signal].
		newFullFileName := self fullName: newName.
		^ self newFileNamed: newFullFileName].

	Abort signal: 'Please close this to abort file opening'! !

!StandardFileStream class methodsFor: 'file creation' stamp: 'TPR 8/13/1999 21:22'!
fileNamed: fileName
	"Open a file with the given name for reading and writing. If the name has no directory part, then the file will be created in the default directory. If the file already exists, its prior contents may be modified or replaced, but the file will not be truncated on close."

	^ self new open: (self fullName: fileName) forWrite: true
! !

!StandardFileStream class methodsFor: 'file creation' stamp: 'SeanDeNigris 7/12/2012 08:44'!
forceNewFileNamed: fileName 
	"Create a new file with the given name, and answer a stream opened 
	for writing on that file. If the file already exists, delete it without 
	asking before creating the new file."
	| dir localName fullName f |
	
	fullName := self fullName: fileName.
	(self isAFileNamed: fullName)
		ifFalse: [f := self new open: fullName forWrite: true.
			^ f ifNil: ["Failed to open the file"
					(FileDoesNotExistException fileName: fullName) signal]].
			
	dir := fullName asFileReference.
	localName := dir basename.
	dir delete.
	dir exists ifTrue: [(CannotDeleteFileException new
			messageText: 'Could not delete the old version of file ' , fullName) signal].
		
	f := self new open: fullName forWrite: true.
	^ f ifNil: ["Failed to open the file"
			(FileDoesNotExistException fileName: fullName) signal].! !

!StandardFileStream class methodsFor: 'file creation' stamp: 'stephaneducasse 2/4/2006 20:32'!
isAFileNamed: fileName
	"Answer true if a file of the given name exists."

	| f |
	f := self new open: fileName forWrite: false.
	f ifNil: [^ false].
	f close.
	^ true
! !

!StandardFileStream class methodsFor: 'file creation' stamp: 'stephaneducasse 2/4/2006 20:32'!
newFileNamed: fileName
 	"Create a new file with the given name, and answer a stream opened for writing on that file. If the file already exists, ask the user what to do."

	| fullName |
	fullName := self fullName: fileName.

	^(self isAFileNamed: fullName)
		ifTrue: ["file already exists:"
			(FileExistsException fileName: fullName fileClass: self) signal]
		ifFalse: [self new open: fullName forWrite: true]

! !

!StandardFileStream class methodsFor: 'file creation' stamp: 'StephaneDucasse 8/30/2009 17:35'!
oldFileNamed: fileName
	"Open an existing file with the given name for reading and writing. If the name has no directory part, then the  default directory will be assumed. If the file does not exist, an exception is signaled. If the file exists, its prior contents may be modified or replaced, but the file will not be truncated on close."

	| fullName |
	fullName := self fullName: fileName.

	^(self isAFileNamed: fullName)
		ifTrue: [self new open: fullName forWrite: true]
		ifFalse: ["File does not exist..."
			(FileDoesNotExistException fileName: fullName) signal]! !

!StandardFileStream class methodsFor: 'error handling' stamp: 'SeanDeNigris 7/12/2012 08:44'!
readOnlyFileDoesNotExistUserHandling: fullFileName

	| dir files choices selection newName fileName |
	
	dir := fullFileName asFileReference.
	files := dir parent fileNames.
	fileName := dir basename.
	
	choices := fileName correctAgainst: files.
	choices add: 'Choose another name'.
	choices add: 'Cancel'.
	
	selection := UIManager default chooseFrom: choices lines: (Array with: 5)
		title: (fullFileName asFileReference basename), ' does not exist.'.
		
	selection = choices size ifTrue:["cancel" ^ nil "should we raise another exception here?"].
	selection < (choices size - 1) ifTrue: [
		newName := dir parent / (choices at: selection)].
	selection = (choices size - 1) ifTrue: [
		newName := UIManager default 
							request: 'Enter a new file name' 
							initialAnswer: fileName].
	newName = '' ifFalse: [^ self readOnlyFileNamed: (self fullName: newName)].
	^ self error: 'Could not open a file'! !

!StandardFileStream class methodsFor: 'file creation' stamp: 'GuillermoPolito 9/1/2010 18:21'!
readOnlyFileNamed: fileName 
	"Open an existing file with the given name for reading."

	| fullName f |
	fullName := self fullName: fileName.
	f := self new open: fullName forWrite: false.
	^ f ifNil: ["File does not exist..."
			((FileDoesNotExistException fileName: fullName) readOnly: true) signal].

	"StandardFileStream readOnlyFileNamed: 'kjsd.txt' "! !

!StandardFileStream class methodsFor: 'registry' stamp: 'HenrikSperreJohansen 5/20/2010 22:38'!
register: anObject
	^self registry add: anObject! !

!StandardFileStream class methodsFor: 'registry' stamp: 'HenrikSperreJohansen 5/20/2010 22:39'!
registry
	^Registry ifNil: [Registry := WeakRegistry new]
! !

!StandardFileStream class methodsFor: 'registry' stamp: 'GuillermoPolito 3/20/2012 22:49'!
retryWithGC: execBlock until: testBlock forFileNamed: fullName
	"Re-implemented to only force GC if a file with the given name exists"
	| blockValue foundIt |
	blockValue := execBlock value.
	(testBlock value: blockValue) ifTrue:[^blockValue].
	"See if we have a file with the given name"
	foundIt := self registry keys "hold on strongly for now" 
		anySatisfy:[:file| file name sameAs: fullName].
	foundIt ifFalse:[^blockValue].
	Smalltalk garbageCollectMost.
	blockValue := execBlock value.
	(testBlock value: blockValue) ifTrue:[^blockValue].
	Smalltalk garbageCollect.
	^execBlock value.! !

!StandardFileStream class methodsFor: 'registry' stamp: 'HenrikSperreJohansen 5/20/2010 22:39'!
unregister: anObject
	^self registry remove: anObject ifAbsent:[]! !

!StandardFileStream methodsFor: 'finalization' stamp: 'ar 3/21/98 18:16'!
actAsExecutor
	super actAsExecutor.
	name := nil.! !

!StandardFileStream methodsFor: 'properties-setting' stamp: 'ul 12/3/2009 02:17'!
ascii
	"opposite of binary"
	buffer1 := String new: 1.
	collection ifNotNil: [ collection := collection asString ]! !

!StandardFileStream methodsFor: 'read, write, position' stamp: 'ul 12/3/2009 06:44'!
atEnd
	"Answer whether the receiver is at its end.  "
	
	collection ifNotNil: [
		position < readLimit ifTrue: [ ^false ] ].
	^self primAtEnd: fileID! !

!StandardFileStream methodsFor: 'read, write, position' stamp: 'ul 12/6/2009 04:10'!
basicNext
	"Answer the next byte from this file, or nil if at the end of the file."
	
	| count |
	<primitive: 65>
	collection ifNotNil: [
		position < readLimit 
			ifFalse: [ 
				readLimit := self primRead: fileID into: collection startingAt: 1 count: collection size.
				position := 0.
				readLimit = 0 ifTrue: [ ^nil ] ].
		^collection at: (position := position + 1) ].	
	count := self primRead: fileID into: buffer1 startingAt: 1 count: 1.
	count = 1
		ifTrue: [ ^buffer1 at: 1 ]
		ifFalse: [ ^nil ]! !

!StandardFileStream methodsFor: 'properties-setting' stamp: 'ul 12/3/2009 02:18'!
binary
	buffer1 := ByteArray new: 1.
	collection ifNotNil: [ collection := collection asByteArray ]! !

!StandardFileStream methodsFor: 'open/close' stamp: 'ul 12/4/2009 07:18'!
close
	"Close this file."

	fileID ifNotNil: [
		collection ifNotNil: [
			readLimit := position := 0 ].
		self primClose: fileID.
		self unregister.
		fileID := nil].
! !

!StandardFileStream methodsFor: 'open/close' stamp: 'ul 12/6/2009 02:45'!
closed
	"Answer true if this file is closed."

	^ fileID == nil or: [ (self primSizeNoError: fileID) == nil ]
! !

!StandardFileStream methodsFor: 'private' stamp: 'nice 11/22/2009 18:03'!
collectionSpecies
	"Answer the species of collection into which the receiver can stream"
	
	^buffer1 species! !

!StandardFileStream methodsFor: 'access' stamp: 'SeanDeNigris 7/12/2012 08:45'!
directory
	"Return the directory containing this file."

	^ self fullName asFileReference parent
! !

!StandardFileStream methodsFor: 'private' stamp: 'ul 12/6/2009 04:12'!
disableReadBuffering

	collection ifNotNil: [
		position < readLimit
			ifTrue: [
				| currentPosition |
				currentPosition := self position.
				collection := readLimit := position := nil.
				self position: currentPosition ]
			ifFalse: [
				collection := readLimit := position := nil ] ]
		! !

!StandardFileStream methodsFor: 'private' stamp: 'ul 12/6/2009 02:49'!
enableReadBuffering

	collection ifNil: [
		buffer1 ifNotNil: [
			collection := self collectionSpecies new: 2048 ] ].
	readLimit := position := 0! !

!StandardFileStream methodsFor: 'open/close' stamp: 'jm 9/21/1998 16:20'!
ensureOpen
	"Make sure that this file really is open."

	self closed ifTrue: [^ self reopen].
	(self primSizeNoError: fileID) ifNotNil: [^ self].
	self reopen.
! !

!StandardFileStream methodsFor: 'access' stamp: 'CamilloBruni 8/1/2012 16:05'!
file
	"Answer the object representing the receiver's file.  Need for compatibility with some calls -- check senders. "

	^ self! !

!StandardFileStream methodsFor: 'finalization' stamp: 'ar 10/7/1998 15:44'!
finalize
	self primCloseNoError: fileID.! !

!StandardFileStream methodsFor: 'read, write, position' stamp: 'stephaneducasse 2/4/2006 20:32'!
findString: string
	"Fast version of #upToAll: to find a String in a file starting from the beginning.
	Returns the position and also sets the position there.
	If string is not found 0 is returned and position is unchanged."

	| pos buffer count oldPos sz |
	oldPos := self position.
	self reset.
	sz := self size.
	pos := 0.
	buffer := String new: 2000.
	[ buffer := self nextInto: buffer.
	(count := buffer findString: string) > 0
		ifTrue: ["Found the string part way into buffer"
			self position: pos.
			self next: count - 1.
			^self position ].
	pos := ((pos + 2000 - string size) min: sz).
	self position: pos.
	pos = sz] whileFalse.
	"Never found it, and hit end of file"
	self position: oldPos.
	^0! !

!StandardFileStream methodsFor: 'read, write, position' stamp: 'stephaneducasse 2/4/2006 20:32'!
findStringFromEnd: string
	"Fast version to find a String in a file starting from the end.
	Returns the position and also sets the position there.
	If string is not found 0 is returned and position is unchanged."

	| pos buffer count oldPos |
	oldPos := self position.
	self setToEnd.
	pos := self position.
	[ pos := ((pos - 2000 + string size) max: 0).  "the [+ string size] allows for the case where the end of the search string is at the beginning of the current buffer"
	self position: pos.
	buffer := self next: 2000.
	(count := buffer findString: string) > 0
		ifTrue: ["Found the string part way into buffer"
			self position: pos.
			self next: count-1.  "use next instead of position:, so that CrLfFileStream can do its magic if it is being used"
			^self position].
	pos = 0] whileFalse.
	"Never found it, and hit beginning of file"
	self position: oldPos.
	^0! !

!StandardFileStream methodsFor: 'read, write, position' stamp: 'ar 2/6/2001 17:59'!
flush
	"Flush pending changes"
	^self primFlush: fileID! !

!StandardFileStream methodsFor: 'private' stamp: 'ul 12/6/2009 04:11'!
flushReadBuffer

	collection ifNotNil: [
		position < readLimit ifTrue: [
			| currentPosition |
			currentPosition := self position.
			position := readLimit := 0.
			self primSetPosition: fileID to: currentPosition ] ]! !

!StandardFileStream methodsFor: 'access' stamp: 'jm 9/21/1998 14:19'!
fullName
	"Answer this file's full path name."

	^ name
! !

!StandardFileStream methodsFor: 'properties-setting' stamp: 'CamilloBruni 5/7/2012 11:36'!
getFileType
	"On the Macintosh, get the file type and creator of this file. On other platforms, do nothing."

	^ self fullName asFileReference macTypeAndCreator
! !

!StandardFileStream methodsFor: 'properties-setting' stamp: 'nice 11/22/2009 18:21'!
insertLineFeeds
	"(FileStream oldFileNamed: 'BBfix2.st') insertLineFeeds"
	| s crLf f |
	crLf := String crlf.
	s := self next: self size.
	self close.
	f := FileStream newFileNamed: self name.
	s linesDo: [:line | f nextPutAll: line; nextPutAll: crLf].
	f close! !

!StandardFileStream methodsFor: 'properties-setting' stamp: ''!
isBinary
	^ buffer1 class == ByteArray! !

!StandardFileStream methodsFor: 'access' stamp: 'CamilloBruni 8/1/2012 16:10'!
isDirectory
	"Answer whether the receiver represents a directory.  For the post-transition case, uncertain what to do."
	^ false! !

!StandardFileStream methodsFor: 'properties-setting' stamp: 'tk 11/4/1998 19:17'!
isReadOnly

	^ rwmode not
! !

!StandardFileStream methodsFor: 'access' stamp: 'CamilloBruni 5/26/2012 12:26'!
localName
	^ name ifNotNil: [
		(name findTokens: FileSystem disk delimiter asString) last]! !

!StandardFileStream methodsFor: 'access' stamp: 'jm 9/21/1998 14:19'!
name
	"Answer this file's full path name."

	^ name
! !

!StandardFileStream methodsFor: 'read, write, position' stamp: 'mir 2/25/2000 12:37'!
next
	"Answer the next byte from this file, or nil if at the end of the file."

	^ self basicNext! !

!StandardFileStream methodsFor: 'read, write, position' stamp: 'CamilloBruni 8/1/2012 16:12'!
next: n
	"Return a string with the next n characters of the filestream in it."
	^ self nextInto: (self collectionSpecies new: n)! !

!StandardFileStream methodsFor: 'read, write, position' stamp: 'HenrikSperreJohansen 3/15/2010 21:34'!
next: n into: aString startingAt: startIndex
	"Read n bytes into the given string.
	Return aString or a partial copy if less than
	n elements have been read."
	|read|
	read := (self readInto: aString startingAt: startIndex count: n).
	^read = n 
		ifTrue: [ aString ]
		ifFalse: [ aString copyFrom: 1 to: startIndex + read - 1 ]	
	! !

!StandardFileStream methodsFor: 'read, write, position' stamp: 'ul 12/6/2009 04:11'!
next: anInteger putAll: aString startingAt: startIndex
	"Store the next anInteger elements from the given collection."
	rwmode ifFalse: [^ self error: 'Cannot write a read-only file'].
	collection ifNotNil: [
		position < readLimit ifTrue: [ self flushReadBuffer ] ].	
	self primWrite: fileID from: aString startingAt: startIndex count: anInteger.
	^aString! !

!StandardFileStream methodsFor: 'read, write, position' stamp: 'ul 12/6/2009 04:11'!
nextPut: char
	"Write the given character to this file."

	rwmode ifFalse: [^ self error: 'Cannot write a read-only file'].
	collection ifNotNil: [ 
		position < readLimit ifTrue: [ self flushReadBuffer ] ].
	buffer1 at: 1 put: char.
	self primWrite: fileID from: buffer1 startingAt: 1 count: 1.
	^ char
! !

!StandardFileStream methodsFor: 'read, write, position' stamp: 'MarianoMartinezPeck 10/26/2010 11:32'!
nextPutAll: stringOrByteArray
	"Write all the characters of the given string to this file."

	rwmode ifFalse: [^ self error: 'Cannot write a read-only file'].
	collection ifNotNil: [ 
		position < readLimit ifTrue: [ self flushReadBuffer ] ].
	self primWrite: fileID from: stringOrByteArray startingAt: 1 count: stringOrByteArray basicSize.
	^ stringOrByteArray
! !

!StandardFileStream methodsFor: 'read, write, position' stamp: 'tk 2/5/2000 21:58'!
nextWordsInto: aBitmap
	"Note: The file primitives automatically adjust for word based objects."

	self next: aBitmap basicSize into: aBitmap startingAt: 1.
	aBitmap restoreEndianness.
	^ aBitmap! !

!StandardFileStream methodsFor: 'open/close' stamp: 'CamilloBruni 8/1/2012 16:12'!
open
	"For compatibility with a few existing things."

	^ self reopen! !

!StandardFileStream methodsFor: 'open/close' stamp: 'CamilloBruni 8/1/2012 16:12'!
open: fileName forWrite: writeMode 
	"Open the file with the given name. If writeMode is true, allow writing, otherwise open the file in read-only mode."
	| f |
	f := fileName asVmPathName.

	fileID := StandardFileStream retryWithGC:[self primOpen: f writable: writeMode] 
					until:[:id| id notNil] 
					forFileNamed: fileName.
	fileID ifNil: [^ nil].  "allows sender to detect failure"
	name := fileName.
	self register.
	rwmode := writeMode.
	buffer1 := String new: 1.
	self enableReadBuffering
	! !

!StandardFileStream methodsFor: 'open/close' stamp: 'StephaneDucasse 5/28/2011 11:43'!
openOnHandle: aFileID name: streamName forWrite: writeMode
	"Initialize the file with the given handle. If writeMode is true then
	 allow writing, otherwise put the file in read-only mode.
	 N.B. Do _not_ register the stream.  We do not want it to be
	 closed implicitly (e.g. on GC).  There may be multiple instances
	 accessing the same stream.  The stream is not a file."
	fileID := aFileID.
	name := streamName.
	rwmode := writeMode.
	buffer1 := String new: 1.
	self enableReadBuffering! !

!StandardFileStream methodsFor: 'open/close' stamp: 'CamilloBruni 8/1/2012 16:12'!
openReadOnly
	"Open the receiver as a read-only file."

	^ self open: name forWrite: false! !

!StandardFileStream methodsFor: 'read, write, position' stamp: 'ul 12/5/2009 07:21'!
padToEndWith: aChar
	"On the Mac, files do not truncate.  One can delete the old file and write a new one, but sometime deletion fails (file still open? file stale?).  This is a sad compromise.  Just let the file be the same length but pad it with a harmless character."

	| pad |
	self atEnd ifTrue: [^ self].
	pad := self isBinary 
		ifTrue: [aChar asCharacter asciiValue]	"ok for char or number"
		ifFalse: [aChar asCharacter].
	self nextPutAll: (self collectionSpecies new: ((self size - self position) min: 20000) 
							withAll: pad).! !

!StandardFileStream methodsFor: 'read, write, position' stamp: 'stephaneducasse 2/4/2006 20:32'!
peek
	"Answer what would be returned if the message next were sent to the receiver. If the receiver is at the end, answer nil.  "
	| next |
	self atEnd ifTrue: [^ nil].
	next := self basicNext.
	self position: self position - 1.
	^ next! !

!StandardFileStream methodsFor: 'access' stamp: 'CamilloBruni 9/5/2012 11:26'!
peekFor: item 
	"Answer false and do not advance if the next element is not equal to item, or if this stream is at the end.  If the next element is equal to item, then advance over it and return true"
	| next |
	(next := self next) == nil ifTrue: [^ false].
	item = next ifTrue: [^ true].
	self skip: -1.
	^ false! !

!StandardFileStream methodsFor: 'read, write, position' stamp: 'tk 10/19/2001 11:29'!
peekLast
	"Return that item just put at the end of the stream"

	^ buffer1 size > 0 
		ifTrue: [buffer1 last]
		ifFalse: [nil]
! !

!StandardFileStream methodsFor: 'read, write, position' stamp: 'CamilloBruni 8/1/2012 16:13'!
position
	"Return the receiver's current file position."

	collection ifNotNil: [
		position < readLimit ifTrue: [
			^(self primGetPosition: fileID) - readLimit + position ] ].
	^self primGetPosition: fileID! !

!StandardFileStream methodsFor: 'read, write, position' stamp: 'CamilloBruni 8/1/2012 16:13'!
position: pos
	"Set the receiver's position as indicated."

	collection ifNotNil: [
		position < readLimit ifTrue: [
			| newPosition |
			newPosition := pos - (self primGetPosition: fileID) + readLimit.
			newPosition < 0 ifTrue: [
					| offsetPos |
					self primSetPosition: fileID to: (offsetPos := pos - (collection size // 4) max: 0).
					readLimit := self primRead: fileID into: collection startingAt: 1 count: collection size.
					position := pos - offsetPos.
					^self ].
			newPosition < readLimit 
				ifTrue: [
					position := newPosition.
					^self ]
				ifFalse: [
					readLimit := position := 0 ] ] ].
	^self primSetPosition: fileID to: pos! !

!StandardFileStream methodsFor: 'primitives' stamp: 'ar 2/2/2001 15:09'!
primAtEnd: id
	"Answer true if the file position is at the end of the file."

	<primitive: 'primitiveFileAtEnd' module: 'FilePlugin'>
	self primitiveFailed
! !

!StandardFileStream methodsFor: 'primitives' stamp: 'ar 2/2/2001 15:09'!
primClose: id
	"Close this file."

	<primitive: 'primitiveFileClose' module: 'FilePlugin'>
	self primitiveFailed
! !

!StandardFileStream methodsFor: 'primitives' stamp: 'ar 2/2/2001 15:09'!
primCloseNoError: id
	"Close this file. Don't raise an error if the primitive fails."

	<primitive: 'primitiveFileClose' module: 'FilePlugin'>
! !

!StandardFileStream methodsFor: 'dnd requests' stamp: 'VeronicaUquillas 6/11/2010 15:32'!
primDropRequestFileHandle: dropIndex
	"Primitive. Return the (read-only) file handle for some file that was just dropped onto the application.
	Fail if dropIndex is out of range or the primitive is not supported."
	<primitive: 'primitiveDropRequestFileHandle' module:'DropPlugin'>
	^nil! !

!StandardFileStream methodsFor: 'dnd requests' stamp: 'VeronicaUquillas 6/11/2010 15:33'!
primDropRequestFileName: dropIndex
	"Primitive. Return the file name for some file that was just dropped onto the application.
	Fail if dropIndex is out of range or the primitive is not supported."
	<primitive: 'primitiveDropRequestFileName' module:'DropPlugin'>
	^nil! !

!StandardFileStream methodsFor: 'primitives' stamp: 'stephaneducasse 2/4/2006 20:32'!
primFlush: id
	"Flush pending changes to the disk"
	| p |
	<primitive: 'primitiveFileFlush' module: 'FilePlugin'>
	"In some OS's seeking to 0 and back will do a flush"
	p := self position.
	self position: 0; position: p! !

!StandardFileStream methodsFor: 'primitives' stamp: 'ar 2/2/2001 15:09'!
primGetPosition: id
	"Get this files current position."

	<primitive: 'primitiveFileGetPosition' module: 'FilePlugin'>
	self primitiveFailed
! !

!StandardFileStream methodsFor: 'primitives' stamp: 'ar 2/2/2001 15:09'!
primOpen: fileName writable: writableFlag
	"Open a file of the given name, and return the file ID obtained.
	If writableFlag is true, then
		if there is none with this name, then create one
		else prepare to overwrite the existing from the beginning
	otherwise
		if the file exists, open it read-only
		else return nil"

	<primitive: 'primitiveFileOpen' module: 'FilePlugin'>
	^ nil
! !

!StandardFileStream methodsFor: 'primitives' stamp: 'ar 2/2/2001 15:09'!
primRead: id into: byteArray startingAt: startIndex count: count
	"Read up to count bytes of data from this file into the given string or byte array starting at the given index. Answer the number of bytes actually read."

	<primitive: 'primitiveFileRead' module: 'FilePlugin'>
	self closed ifTrue: [^ self error: 'File is closed'].
	self error: 'File read failed'.
! !

!StandardFileStream methodsFor: 'primitives' stamp: 'ar 2/2/2001 15:09'!
primSetPosition: id to: anInteger
	"Set this file to the given position."

	<primitive: 'primitiveFileSetPosition' module: 'FilePlugin'>
	self primitiveFailed
! !

!StandardFileStream methodsFor: 'primitives' stamp: 'ar 2/2/2001 15:09'!
primSize: id
	"Answer the size of this file."

	<primitive: 'primitiveFileSize' module: 'FilePlugin'>
	self primitiveFailed
! !

!StandardFileStream methodsFor: 'primitives' stamp: 'ar 2/2/2001 15:09'!
primSizeNoError: id
	"Answer the size of this file. Answer nil if the primitive fails; this indicates that the file handle has become stale."

	<primitive: 'primitiveFileSize' module: 'FilePlugin'>
	^ nil
! !

!StandardFileStream methodsFor: 'primitives' stamp: 'JMM 5/24/2001 21:55'!
primTruncate: id to: anInteger
	"Truncate this file to the given position."

	<primitive: 'primitiveFileTruncate' module: 'FilePlugin'>
	self primitiveFailed
! !

!StandardFileStream methodsFor: 'primitives' stamp: 'MarianoMartinezPeck 10/26/2010 11:32'!
primWrite: id from: stringOrByteArray startingAt: startIndex count: count
	"Write count bytes onto this file from the given string or byte array starting at the given index. Answer the number of bytes written."

	<primitive: 'primitiveFileWrite' module: 'FilePlugin'>
	(FileWriteError fileName: name)
		signal: (self closed
			ifTrue: [ 'File ', name, ' is closed' ]
			ifFalse: [ 'File ', name, ' write failed' ])! !

!StandardFileStream methodsFor: 'access' stamp: 'CamilloBruni 8/1/2012 16:13'!
printOn: aStream
	"Put a printed version of the receiver onto aStream."

	aStream nextPutAll: self class name; nextPutAll: ': '; print: name! !

!StandardFileStream methodsFor: 'read, write, position' stamp: 'HenrikSperreJohansen 3/15/2010 23:06'!
readInto: byteArray startingAt: startIndex count: count
	"Read up to n bytes into the given byteArray.
	Return the number of bytes actually read."
	
	| newN newStartIndex available|
	available := 0.
	collection 
		ifNil: [ 
			newN := count.
			newStartIndex := startIndex ]
		ifNotNil: [
			byteArray class isBytes 
				ifFalse: [ 
					position < readLimit ifTrue: [ self flushReadBuffer ].
					newN := count.
					newStartIndex := startIndex ]
				ifTrue: [
					(available := readLimit - position) > 0 
						ifTrue: [
							| bufferedCount |
							bufferedCount := count min: available.
							byteArray
								replaceFrom: startIndex
								to: startIndex + bufferedCount - 1
								with: collection
								startingAt: position + 1.
							position := position + bufferedCount.
							bufferedCount = count ifTrue: [ ^count ] ].
					newN := count - available.
					newStartIndex := startIndex + available ] ].
	^available + (self primRead: fileID into: byteArray
				startingAt: newStartIndex count: newN).
! !

!StandardFileStream methodsFor: 'properties-setting' stamp: 'ul 12/3/2009 06:39'!
readOnly
	"Make this file read-only."

	rwmode := false.
! !

!StandardFileStream methodsFor: 'read, write, position' stamp: 'yo 10/31/2002 22:33'!
readOnlyCopy

	^ self class readOnlyFileNamed: self name.
! !

!StandardFileStream methodsFor: 'properties-setting' stamp: 'ul 12/3/2009 06:39'!
readWrite
	"Make this file writable."

	rwmode := true.
! !

!StandardFileStream methodsFor: 'registry' stamp: 'ar 3/21/98 17:23'!
register
	^self class register: self! !

!StandardFileStream methodsFor: 'open/close' stamp: 'ul 12/6/2009 04:11'!
reopen
	"Close and reopen this file. The file position is reset to zero."
	"Details: Files that were open when a snapshot occurs are no longer valid when the snapshot is resumed. This operation re-opens the file if that has happened."

	fileID ifNotNil: [
		collection ifNotNil: [
			position < readLimit ifTrue: [
				self flushReadBuffer ] ].
		self primCloseNoError: fileID ].
	self open: name forWrite: rwmode.
! !

!StandardFileStream methodsFor: 'dnd requests' stamp: 'VeronicaUquillas 6/11/2010 15:33'!
requestDropStream: dropIndex
	"Return a read-only stream for some file the user has just dropped onto the application."
	| rawName |
	rawName := self primDropRequestFileName: dropIndex.
	name :=  (FilePath pathName: rawName isEncoded: true) asSqueakPathName.
	fileID := self primDropRequestFileHandle: dropIndex.
	fileID == nil ifTrue:[^nil].
	self register.
	rwmode := false.
	buffer1 := String new: 1.
	self enableReadBuffering
! !

!StandardFileStream methodsFor: 'access' stamp: 'ar 6/16/2002 18:58'!
reset
	self ensureOpen.
	self position: 0.! !

!StandardFileStream methodsFor: 'properties-setting' stamp: 'SeanDeNigris 7/12/2012 08:44'!
setFileTypeToObject
	"On the Macintosh, set the file type and creator of this file to be a Pharo object file. On other platforms, do nothing. Setting the file type allows Pharo object files to be sent as email attachments and launched by double-clicking. On other platforms, similar behavior is achieved by creating the file with the '.sqo' file name extension."

	self fullName asFileReference setMacType: 'SOBJ' creator: 'FAST'.
! !

!StandardFileStream methodsFor: 'read, write, position' stamp: 'CamilloBruni 8/1/2012 16:16'!
setToEnd
	"Set the position of the receiver to the end of file."

	self position: self size! !

!StandardFileStream methodsFor: 'access' stamp: 'CamilloBruni 8/1/2012 16:16'!
size
	"Answer the size of the file in characters."

	^ self primSize: fileID! !

!StandardFileStream methodsFor: 'read, write, position' stamp: 'CamilloBruni 8/1/2012 16:16'!
skip: n
	"Set the character position to n characters from the current position.
	Error if not enough characters left in the file."

	collection ifNotNil: [
		position < readLimit ifTrue: [
			| newPosition |
			((newPosition := position + n) >= 0 and: [ newPosition < readLimit ])
				ifTrue: [ 
					position := newPosition.
					^self ] ] ].
	self position: self position + n! !

!StandardFileStream methodsFor: 'read, write, position' stamp: 'JMM 5/24/2001 22:00'!
truncate
	"Truncate to zero"

	^ self truncate: 0! !

!StandardFileStream methodsFor: 'read, write, position' stamp: 'JMM 5/24/2001 22:47'!
truncate: pos
	"Truncate to this position"

	self position: pos.
	^self primTruncate: fileID to: pos! !

!StandardFileStream methodsFor: 'registry' stamp: 'ar 3/21/98 17:23'!
unregister
	^self class unregister: self! !

!StandardFileStream methodsFor: 'read, write, position' stamp: 'ul 1/15/2010 20:16'!
upTo: delim 
	"Fast version to speed up nextChunk"
	| pos buffer count |
	collection ifNotNil: [
		(position < readLimit and: [
			(pos := collection indexOf: delim startingAt: position + 1) <= readLimit and: [
				pos > 0 ] ]) ifTrue: [
					^collection copyFrom: position + 1 to: (position := pos) - 1 ] ].
	pos := self position.
	buffer := self next: 2000.
	(count := buffer indexOf: delim) > 0 ifTrue: 
		["Found the delimiter part way into buffer"
		self position: pos + count.
		^ buffer copyFrom: 1 to: count - 1].
	self atEnd ifTrue:
		["Never found it, and hit end of file"
		^ buffer].
	"Never found it, but there's more..."
	^ buffer , (self upTo: delim)! !

!StandardFileStream methodsFor: 'read, write, position' stamp: 'nice 12/8/2009 19:03'!
upToAnyOf: delimiters do: aBlock
	"Fast version to speed up nextChunk"
	| pos buffer count result |
	collection ifNotNil: [
		(position < readLimit and: [
			(count := collection indexOfAnyOf: delimiters startingAt: position + 1) <= readLimit and: [
				count > 0 ] ]) ifTrue: [
					result := collection copyFrom: position + 1 to: (position := count) - 1.
					aBlock value: (collection at: count).
					^result ] ].
	pos := self position.
	buffer := self next: 2000.
	(count := buffer indexOfAnyOf: delimiters) > 0 ifTrue: 
		["Found one of the delimiters part way into buffer"
		self position: pos + count.
		aBlock value: (buffer at: count).
		^ buffer copyFrom: 1 to: count - 1].
	self atEnd ifTrue:
		["Never found it, and hit end of file"
		^ buffer].
	"Never found it, but there's more..."
	^ buffer , (self upToAnyOf: delimiters do: aBlock)! !

!StandardFileStream methodsFor: 'read, write, position' stamp: 'ul 12/3/2009 06:42'!
upToEnd
	"Answer a subcollection from the current access position through the last element of the receiver."

	^self collectionSpecies streamContents: [ :newStream |
		| next |
		[ (next := self next) == nil ] whileFalse: [
			newStream nextPut: next ] ]! !

!String methodsFor: '*files' stamp: 'CamilloBruni 5/10/2012 16:00'!
asFileName
	"Answer a String made up from the receiver that is an acceptable file 
	name."

	| string checkedString |
	string := FileSystem disk checkName: self fixErrors: true.
	checkedString := (FilePath pathName: string) asVmPathName.
	^ (FilePath pathName: checkedString isEncoded: true) asSqueakPathName.
! !

!String methodsFor: '*files' stamp: 'yo 2/24/2005 18:33'!
asVmPathName

	^ (FilePath pathName: self) asVmPathName.
! !

!MIMEDocument methodsFor: '*files' stamp: 'MarcusDenker 3/21/2011 14:34'!
saveToFile: anAbsolutePathString

	FileStream forceNewFileNamed: anAbsolutePathString do: [ :str |
		str binary.
		str nextPutAll: (self contents) ].! !

!Integer methodsFor: '*Files' stamp: 'cmm 2/15/2010 15:52'!
asBytesDescription
	"Answer a terse, easily-readable representation of this Integer reprsenting a number of bytes.  Useful for file-browsers."
	| suffixes |
	suffixes := { 'k'"ilobytes". 'M'"egabytes". 'G'"igabytes". 'T'"erabytes". 'P'"etabytes". 'E'"xabytes". 'Z'"ettabytes". 'Y'"ottabytes"}.
	suffixes size to: 1 by: -1 do:
		[ : index |  | units |
		units := 1000 raisedTo: index.
		self > units ifTrue: [ ^ ((self / units) asFloat roundTo: 0.01) asString, (suffixes at: index) ] ].
	^ self asString! !

Error subclass: #FileStreamException
	instanceVariableNames: 'fileName'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Files-Kernel'!

FileStreamException subclass: #CannotDeleteFileException
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Files-Kernel'!

FileStreamException subclass: #FileDoesNotExistException
	instanceVariableNames: 'readOnly'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Files-Kernel'!

!FileDoesNotExistException class methodsFor: 'examples' stamp: 'mir 2/29/2000 11:44'!
example
	"FileDoesNotExistException example"

	| result |
	result := [(StandardFileStream readOnlyFileNamed: 'error42.log') contentsOfEntireFile]
		on: FileDoesNotExistException
		do: [:ex | 'No error log'].
	Transcript show: result; cr! !

!FileDoesNotExistException methodsFor: 'exceptiondescription' stamp: 'pavel.krivanek 6/17/2010 10:00'!
defaultAction
	"The default action taken if the exception is signaled."

	^ UIManager default fileDoesNotExistsDefaultAction: self
! !

!FileDoesNotExistException methodsFor: 'accessing' stamp: 'mir 7/25/2000 16:41'!
readOnly
	^readOnly == true! !

!FileDoesNotExistException methodsFor: 'accessing' stamp: 'mir 7/25/2000 16:40'!
readOnly: aBoolean
	readOnly := aBoolean! !

FileStreamException subclass: #FileExistsException
	instanceVariableNames: 'fileClass'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Files-Kernel'!

!FileExistsException class methodsFor: 'exceptioninstantiator' stamp: 'LC 10/24/2001 21:50'!
fileName: aFileName fileClass: aClass 
	^ self new
		fileName: aFileName;
		fileClass: aClass! !

!FileExistsException methodsFor: 'exceptiondescription' stamp: 'pavel.krivanek 6/17/2010 10:00'!
defaultAction
	"The default action taken if the exception is signaled."

	^ UIManager default fileExistsDefaultAction: self
! !

!FileExistsException methodsFor: 'accessing' stamp: 'LC 10/24/2001 21:49'!
fileClass
	^ fileClass ifNil: [StandardFileStream]! !

!FileExistsException methodsFor: 'accessing' stamp: 'LC 10/24/2001 21:42'!
fileClass: aClass
	fileClass := aClass! !

!FileStreamException class methodsFor: 'exceptioninstantiator' stamp: 'mir 2/23/2000 20:12'!
fileName: aFileName
	^self new fileName: aFileName! !

!FileStreamException methodsFor: 'exceptiondescription' stamp: 'mir 2/25/2000 17:29'!
fileName
	^fileName! !

!FileStreamException methodsFor: 'exceptionbuilder' stamp: 'mir 2/23/2000 20:13'!
fileName: aFileName
	fileName := aFileName! !

!FileStreamException methodsFor: 'exceptiondescription' stamp: 'mir 2/23/2000 20:13'!
isResumable
	"Determine whether an exception is resumable."

	^true! !

!FileStreamException methodsFor: 'exceptiondescription' stamp: 'mir 2/23/2000 20:14'!
messageText
	
	"Return an exception's message text."

	^messageText == nil
		ifTrue: [fileName printString]
		ifFalse: [messageText]! !

FileStreamException subclass: #FileWriteError
	instanceVariableNames: ''
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Files-Kernel'!

Error subclass: #InvalidDirectoryError
	instanceVariableNames: 'pathName'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Files-Directories'!

!InvalidDirectoryError class methodsFor: 'exceptioninstantiator' stamp: 'ar 5/30/2001 20:49'!
pathName: badPathName
	^self new pathName: badPathName! !

!InvalidDirectoryError methodsFor: 'exceptiondescription' stamp: 'StephaneDucasse 8/30/2009 16:54'!
defaultAction

	 ^#()! !

!InvalidDirectoryError methodsFor: 'accessing' stamp: 'ar 5/30/2001 20:44'!
pathName
	^pathName! !

!InvalidDirectoryError methodsFor: 'accessing' stamp: 'ar 5/30/2001 20:45'!
pathName: badPathName
	pathName := badPathName! !

Object subclass: #AsyncFile
	instanceVariableNames: 'name writeable semaphore fileHandle'
	classVariableNames: 'Busy ErrorCode'
	poolDictionaries: ''
	category: 'Files-Kernel'!

!AsyncFile commentStamp: 'HenrikSperreJohansen 2/16/2012 11:40' prior: 0!
An asynchronous file allows simple file read and write operations to be performed in parallel with other processing. This is useful in multimedia applications that need to stream large amounts of sound or image data from or to a file while doing other work.

Closing the file after its use is currently required to not leak external semaphores. 
!

!AsyncFile class methodsFor: 'initialization' stamp: 'bootstrap 5/31/2006 20:45'!
initialize
	"AsyncFile initialize"

	"Possible abnormal I/O completion results."
	Busy := -1.
	ErrorCode := -2.
! !

!AsyncFile methodsFor: 'as yet unclassified' stamp: 'stephaneducasse 2/4/2006 20:31'!
close

	fileHandle ifNil: [^ self].  "already closed"
	self primClose: fileHandle.
	Smalltalk unregisterExternalObject: semaphore.
	semaphore := nil.
	fileHandle := nil.
! !

!AsyncFile methodsFor: 'as yet unclassified' stamp: 'di 7/6/1998 10:58'!
fileHandle
	^ fileHandle! !

!AsyncFile methodsFor: 'as yet unclassified' stamp: 'ar 6/3/2007 22:13'!
open: fullFileName forWrite: aBoolean
	"Open a file of the given name, and return a handle for that file. Answer the receiver if the primitive succeeds, nil otherwise.
	If openForWrite is true, then:
		if there is no existing file with this name, then create one
		else open the existing file in read-write mode
	otherwise:
		if there is an existing file with this name, then open it read-only
		else answer nil."
	"Note: if an exisiting file is opened for writing, it is NOT truncated. If truncation is desired, the file should be deleted before being opened as an asynchronous file."
	"Note: On some platforms (e.g., Mac), a file can only have one writer at a time."

	| semaIndex |
	name := fullFileName.
	writeable := aBoolean.
	semaphore := Semaphore new.
	semaIndex := Smalltalk registerExternalObject: semaphore.
	fileHandle := self primOpen: name asVmPathName forWrite: writeable semaIndex: semaIndex.
	fileHandle ifNil: [
		Smalltalk unregisterExternalObject: semaphore.
		semaphore := nil.
		^ nil].
! !

!AsyncFile methodsFor: 'primitives' stamp: 'ar 2/2/2001 15:09'!
primClose: fHandle
	"Close this file. Do nothing if primitive fails."

	<primitive: 'primitiveAsyncFileClose' module: 'AsynchFilePlugin'>
! !

!AsyncFile methodsFor: 'primitives' stamp: 'ar 2/2/2001 15:09'!
primOpen: fileName forWrite: openForWrite semaIndex: semaIndex
	"Open a file of the given name, and return a handle for that file. Answer the receiver if the primitive succeeds, nil otherwise."

	<primitive: 'primitiveAsyncFileOpen' module: 'AsynchFilePlugin'>
	^ nil
! !

!AsyncFile methodsFor: 'primitives' stamp: 'ar 2/2/2001 15:09'!
primReadResult: fHandle intoBuffer: buffer at: startIndex count: count
	"Copy the result of the last read operation into the given buffer starting at the given index. The buffer may be any sort of bytes or words object, excluding CompiledMethods. Answer the number of bytes read. A negative result means:
		-1 the last operation is still in progress
		-2 the last operation encountered an error"

	<primitive: 'primitiveAsyncFileReadResult' module: 'AsynchFilePlugin'>
	self primitiveFailed
! !

!AsyncFile methodsFor: 'primitives' stamp: 'ar 2/2/2001 15:09'!
primReadStart: fHandle fPosition: fPosition count: count
	"Start a read operation of count bytes starting at the given offset in the given file."

	<primitive: 'primitiveAsyncFileReadStart' module: 'AsynchFilePlugin'>
	self error: 'READ THE COMMENT FOR THIS METHOD.'

"NOTE: This method will fail if there is insufficient C heap to allocate an internal buffer of the required size (the value of count).  If you are trying to read a movie file, then the buffer size will be height*width*2 bytes.  Each Squeak image retains a value to be used for this allocation, and it it initially set to 0.  If you are wish to play a 640x480 movie, you need room for a buffer of 640*480*2 = 614400 bytes.  You should execute the following...

	Smalltalk extraVMMemory 2555000.

Then save-and-quit, restart, and try to open the movie file again.  If you are using Async files in another way, find out the value of count when this failure occurs (call it NNNN), and instead of the above, execute...

	Smalltalk extraVMMemory: Smalltalk extraVMMemory + NNNN

then save-and-quit, restart, and try again.
"

! !

!AsyncFile methodsFor: 'primitives' stamp: 'ar 2/2/2001 15:09'!
primWriteResult: fHandle
	"Answer the number of bytes written. A negative result means:
		-1 the last operation is still in progress
		-2 the last operation encountered an error"

	<primitive: 'primitiveAsyncFileWriteResult' module: 'AsynchFilePlugin'>
	self primitiveFailed
! !

!AsyncFile methodsFor: 'primitives' stamp: 'ar 2/2/2001 15:09'!
primWriteStart: fHandle fPosition: fPosition fromBuffer: buffer at: startIndex count: count
	"Start a write operation of count bytes starting at the given index in the given buffer. The buffer may be any sort of bytes or words object, excluding CompiledMethods. The contents of the buffer are copied into an internal buffer immediately, so the buffer can be reused after the write operation has been started. Fail if there is insufficient C heap to allocate an internal buffer of the requested size."

	<primitive: 'primitiveAsyncFileWriteStart' module: 'AsynchFilePlugin'>
	writeable ifFalse: [^ self error: 'attempt to write a file opened read-only'].
	self primitiveFailed
! !

!AsyncFile methodsFor: 'as yet unclassified' stamp: 'nice 4/16/2009 10:02'!
readByteCount: byteCount fromFilePosition: fPosition onCompletionDo: aBlock
	"Start a read operation to read byteCount's from the given position in this file. and fork a process to await its completion. When the operation completes, evaluate the given block. Note that, since the completion block may run asynchronous, the client may need to use a SharedQueue or a semaphore for synchronization."

	| buffer |
	buffer := String new: byteCount.
	self primReadStart: fileHandle fPosition: fPosition count: byteCount.
	"here's the process that awaits the results:"
	[| n |
		[	semaphore wait.
		  	n := self primReadResult: fileHandle intoBuffer: buffer at: 1 count: byteCount.
		  	n = Busy.
		] whileTrue.  "loop while busy in case the semaphore had excess signals"
		n = ErrorCode ifTrue: [^ self error: 'asynchronous read operation failed'].
		aBlock value: buffer.
	] forkAt: Processor userInterruptPriority.
! !

!AsyncFile methodsFor: 'as yet unclassified' stamp: 'CamilloBruni 5/4/2012 21:16'!
test: byteCount fileName: fileName
	"AsyncFile new test: 10000 fileName: 'testData'"

	| buf1 buf2 bytesWritten bytesRead |
	buf1 := String new: byteCount withAll: $x.
	buf2 := String new: byteCount.
	self open: fileName  asFileReference fullName forWrite: true.
	self primWriteStart: fileHandle
		fPosition: 0
		fromBuffer: buf1
		at: 1
		count: byteCount.
	semaphore wait.
	bytesWritten := self primWriteResult: fileHandle.
	self close.
	
	self open: fileName asFileReference fullName forWrite: false.
	self primReadStart: fileHandle fPosition: 0 count: byteCount.
	semaphore wait.
	bytesRead :=
		self primReadResult: fileHandle
			intoBuffer: buf2
			at: 1
			count: byteCount.
	self close.

	buf1 = buf2 ifFalse: [self error: 'buffers do not match'].
	^ 'wrote ', bytesWritten printString, ' bytes; ',
	   'read ', bytesRead printString, ' bytes'
! !

!AsyncFile methodsFor: 'as yet unclassified' stamp: 'di 7/6/1998 10:58'!
waitForCompletion
	semaphore wait! !

!AsyncFile methodsFor: 'as yet unclassified' stamp: 'nice 4/16/2009 10:02'!
writeBuffer: buffer atFilePosition: fPosition onCompletionDo: aBlock
	"Start an operation to write the contents of the buffer at given position in this file, and fork a process to await its completion. When the write completes, evaluate the given block. Note that, since the completion block runs asynchronously, the client may need to use a SharedQueue or a semaphore for synchronization."

	self primWriteStart: fileHandle
		fPosition: fPosition
		fromBuffer: buffer
		at: 1
		count: buffer size.
	"here's the process that awaits the results:"
	[| n |
		[	semaphore wait.
		  	n := self primWriteResult: fileHandle.
		  	n = Busy.
		] whileTrue.  "loop while busy in case the semaphore had excess signals"
		n = ErrorCode ifTrue: [^ self error: 'asynchronous write operation failed'].
		n = buffer size ifFalse: [^ self error: 'did not write the entire buffer'].
		aBlock value.
	] forkAt: Processor userInterruptPriority.
! !

Object subclass: #FilePath
	instanceVariableNames: 'squeakPathName vmPathName'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Files-Directories'!

!FilePath commentStamp: 'yo 10/19/2004 21:36' prior: 0!
This class absorb the difference of internal and external representation of the file path.  The idea is to keep the internal one as much as possible, and only when it goes to a primitive, the encoded file path, i.e. the native platform representation is passsed to the primitive.

	The converter used is obtained by "LanguageEnvironment defaultFileNameConverter".
!

!FilePath class methodsFor: 'as yet unclassified' stamp: 'yo 2/24/2005 18:38'!
classVersion

	^ 1.
! !

!FilePath class methodsFor: 'instance creation' stamp: 'yo 12/19/2003 16:30'!
pathName: pathName

	^ self pathName: pathName isEncoded: false.
! !

!FilePath class methodsFor: 'instance creation' stamp: 'yo 12/19/2003 16:30'!
pathName: pathName isEncoded: aBoolean

	^ (self new) pathName: pathName isEncoded: aBoolean; yourself.
! !

!FilePath methodsFor: 'conversion' stamp: 'yo 12/19/2003 21:10'!
asSqueakPathName

	^ self pathName.
! !

!FilePath methodsFor: 'conversion' stamp: 'ar 1/31/2005 11:16'!
asString
	^self asSqueakPathName! !

!FilePath methodsFor: 'conversion' stamp: 'yo 2/24/2005 18:45'!
asVmPathName

	^ vmPathName.
! !

!FilePath methodsFor: 'file in/out' stamp: 'stephaneducasse 2/4/2006 20:31'!
copySystemToVm

	(self class instVarNames includes: 'systemPathName') ifTrue: [
		vmPathName := self instVarNamed: 'systemPathName'.
	].

! !

!FilePath methodsFor: 'testing' stamp: 'tpr 11/5/2004 11:39'!
isNullPath
	"an empty path is used to represent the root path(s) when calling the primitive to list directory entries. Some users need to check for this and this is cleaner than grabbing the pathname and assuming it is a plain String"
	^self pathName isEmpty! !

!FilePath methodsFor: 'conversion' stamp: 'yo 12/19/2003 21:07'!
pathName

	^ squeakPathName.
! !

!FilePath methodsFor: 'conversion' stamp: 'CamilloBruni 9/22/2012 10:24'!
pathName: p isEncoded: isEncoded
	| encodingConverter |
	encodingConverter := LanguageEnvironment defaultFileNameConverter.
	isEncoded ifTrue: [
		squeakPathName := p convertFromWithConverter: encodingConverter.
		vmPathName := p.
	] ifFalse: [
		squeakPathName := p isOctetString ifTrue: [p asOctetString] ifFalse: [p].
		vmPathName := squeakPathName convertToWithConverter: encodingConverter.
	].
! !

!FilePath methodsFor: 'conversion' stamp: 'yo 12/19/2003 21:07'!
printOn: aStream

	aStream nextPutAll: 'FilePath('''.
	aStream nextPutAll: squeakPathName.
	aStream nextPutAll: ''')'.
! !

Object subclass: #RemoteString
	instanceVariableNames: 'sourceFileNumber filePositionHi'
	classVariableNames: ''
	poolDictionaries: ''
	category: 'Files-System'!

!RemoteString commentStamp: '<historical>' prior: 0!
My instances provide an external file reference to a piece of text.  It may be the sourceCode of a method, or the class comments of a class.

The changes file or file-in file usually has a chunk that is just the source string of a method:

max: aNumber
	^ self > aNumber ifTrue: [self] ifFalse: [aNumber]!!

I can return either a String or a Text.  Some a chunk is followed by a second chunk (beginning with ]style[) containing style information.  The encoding is like this:

max: aNumber
	^ self > aNumber ifTrue: [self] ifFalse: [aNumber]!!
]style[(14 50 312)f1,f1b,f1LInteger +;i!!

Allowed TextAttributes are TextFontChange, TextEmphasis, TextColor, TextDoIt, TextKern, TextLink, TextURL.  TextFontReference and TextAnchor are not supported.

See PositionableStream nextChunkText and RunArray class scanFrom:.!

!RemoteString class methodsFor: 'as yet unclassified' stamp: ''!
newFileNumber: sourceIndex position: anInteger 
	"Answer an instance of me fora file indexed by sourceIndex, at the 
	position anInteger. Assume that the string is already stored on the file 
	and the instance will be used to access it."

	^self new fileNumber: sourceIndex position: anInteger! !

!RemoteString class methodsFor: 'as yet unclassified' stamp: ''!
newString: aString onFileNumber: sourceIndex 
	"Answer an instance of me for string, aString, on file indexed by 
	sourceIndex. Put the string on the file and create the remote reference."

	^self new string: aString onFileNumber: sourceIndex! !

!RemoteString class methodsFor: 'as yet unclassified' stamp: ''!
newString: aString onFileNumber: sourceIndex toFile: aFileStream
	"Answer an instance of me for string, aString, on file indexed by 
	sourceIndex. Put the string on the file, aFileStream, and create the 
	remote reference. Assume that the index corresponds properly to 
	aFileStream."

	^self new string: aString onFileNumber: sourceIndex toFile: aFileStream! !

!RemoteString methodsFor: 'private' stamp: 'jcg 9/1/2009 00:28'!
checkSum: aString
	"Construct a checksum of the string.  A three byte number represented as Base64 characters."
	| sum shift bytes |
	sum := aString size.
	shift := 0.
	aString do: [:char |
		(shift := shift + 7) > 16 ifTrue: [shift := shift - 17].
			"shift by 7 to keep a change of adjacent chars from xoring to same value"
		sum := sum bitXor: (char asInteger bitShift: shift)
	].
	bytes := ByteArray new: 3.
	sum := sum + 16r10000000000.
	1 to: 3 do: [:ind | bytes at: ind put: (sum digitAt: ind)].
	^bytes base64Encoded! !

!RemoteString methodsFor: 'private' stamp: 'stephaneducasse 2/4/2006 20:32'!
fileNumber: fileNumber position: position 

	sourceFileNumber := fileNumber.
	filePositionHi := position! !

!RemoteString methodsFor: 'accessing' stamp: 'MarianoMartinezPeck 10/29/2011 19:33'!
fileStream 
	"Answer the file stream with position set at the beginning of my string.
	Answer a read only copy to avoid syntax errors when accessed via
	multiple processes."

	| theFile |
	(sourceFileNumber == nil or: [(SourceFiles at: sourceFileNumber) == nil]) ifTrue: [^ nil].
	theFile := (SourceFiles at: sourceFileNumber) readOnlyCopy.
	filePositionHi > theFile size ifTrue: [
		self error: 'RemoteString past end of file' ].
	theFile position: filePositionHi.
	^ theFile! !

!RemoteString methodsFor: 'accessing' stamp: 'Sd 1/28/2011 14:36'!
last
	^self string ifNotNil: [ :s | s last ] ! !

!RemoteString methodsFor: 'accessing' stamp: 'Sd 1/28/2011 14:36'!
position 
	"Answer the location of the string on a file."

	^ filePositionHi ! !

!RemoteString methodsFor: 'accessing' stamp: 'stephaneducasse 2/4/2006 20:31'!
setSourcePointer: aSourcePointer
	sourceFileNumber := SourceFiles fileIndexFromSourcePointer: aSourcePointer.
	filePositionHi := SourceFiles filePositionFromSourcePointer: aSourcePointer! !

!RemoteString methodsFor: 'accessing' stamp: ''!
sourceFileNumber
	"Answer the index of the file on which the string is stored."

	^sourceFileNumber! !

!RemoteString methodsFor: 'accessing' stamp: 'hmm 4/26/2000 20:47'!
sourcePointer
	sourceFileNumber ifNil: [^ 0].
	^SourceFiles sourcePointerFromFileIndex: sourceFileNumber andPosition: filePositionHi! !

!RemoteString methodsFor: 'accessing' stamp: 'MarianoMartinezPeck 10/29/2011 19:33'!
string 
	"Answer the receiver's string if remote files are enabled.
	Use a read only copy to avoid syntax errors when accessed via
	multiple processes."
	
	| theFile |
	(sourceFileNumber == nil or: [(SourceFiles at: sourceFileNumber) == nil]) ifTrue: [^''].
	theFile := (SourceFiles at: sourceFileNumber) readOnlyCopy.
	^[filePositionHi > theFile size ifTrue: [
		self error: 'RemoteString past end of file' ].
		theFile position: filePositionHi.
		theFile nextChunk] ensure: [theFile close] 
	! !

!RemoteString methodsFor: 'private' stamp: 'stephaneducasse 2/4/2006 20:32'!
string: aString onFileNumber: fileNumber
	"Store this as my string if source files exist."
	| theFile |
	(SourceFiles at: fileNumber) == nil ifFalse: 
		[theFile := SourceFiles at: fileNumber.
		theFile setToEnd; cr.
		self string: aString onFileNumber: fileNumber toFile: theFile]! !

!RemoteString methodsFor: 'chunk' stamp: 'StephaneDucasse 9/12/2010 09:53'!
string: aString onFileNumber: fileNumber toFile: aFileStream 
	"Store this as the receiver's string if source files exist."

	| position |
	position := aFileStream position.
	self fileNumber: fileNumber position: position.
	aFileStream nextChunkPut: aString.
	^ self ! !

!RemoteString methodsFor: 'accessing' stamp: 'Sd 1/28/2011 14:44'!
text 
	"Answer the receiver's string asText if remote files are enabled.
	Use a read only copy to avoid syntax errors when accessed via
	multiple processes."
	
	"| theFile |
	self haltOnce.
	(sourceFileNumber == nil or: [(SourceFiles at: sourceFileNumber) == nil]) ifTrue: [^ nil].
	theFile := (SourceFiles at: sourceFileNumber) readOnlyCopy.
	^[theFile position: filePositionHi.
	theFile position > theFile size ifTrue: [
		self error: 'RemoteString past end of file' ].
	theFile nextChunkText string ] ensure: [theFile close]"
	"Code saved should only be ascii and we do not want to have the encoding at the end of the chunk
	this is why we extract the string and convert instead of invoking nextChunkText"
	
	^ self string asText ! !
AsyncFile initialize!
FileStream initialize!
MultiByteFileStream initialize!
